from __future__ import annotations
from typing import List

from abc import ABC, abstractmethod
import numpy.typing as npt
import numpy as np

class TSPreprocessorInterface(ABC):
    @abstractmethod
    def fit(self, 
            X: npt.NDArray[np.float64],
            *args, 
            **kwargs
            ) -> 'TSPreprocessorInterface':
       """
        Fit method of transformer. 

        Parameters
        ----------
        X: ndarray. 
            Input time series data. Should be a 3 dimensional array in TNF fromat.
        args: 
            any other positional arguments for fit method.
        kwargs:
            any keyword argument for fit method 

        Returns
        ----------
        self
            the fitted transformer object
        """
       
       raise NotImplementedError
    
    @abstractmethod
    def transform(self, 
                X: npt.NDArray[np.float64], 
                *args, 
                **kwargs
                ) -> npt.NDArray[np.float64]:
        """
        Transform method. 

        Parameters
        ----------
        X: ndarray
            Input time series data. Should be a 3 dimensional array in TNF fromat.
        args: 
            any other positional arguments for fit method.
        kwargs:
            any keyword argument for fit method.

        Returns
        ----------
        numpy array
            The transformed data as numpy array of dimension T x N x F
        """

        raise NotImplementedError

    @abstractmethod
    def inverse_transform(self, 
            X: npt.NDArray[np.float64], 
            *args, 
            **kwargs
            ) -> npt.NDArray[np.float64]:
        """
        Inverse transform method. 

        Parameters
        ----------
        X: ndarray 
            Input time series data. Should be a 3 dimensional array in TNF fromat.
        args: 
            any other positional arguments for fit method.
        kwargs:
            any keyword argument for fit method.

        Returns
        ----------
        numpy array
            The inverse transform of the input data as numpy array of dimension T x N x F
        """
        
        raise NotImplementedError


from __future__ import annotations
from typing import List

import numpy as np
import numpy.typing as npt
import pandas as pd

from tscluster.preprocessing.utils import broadcast_data

def inertia(
        X: npt.NDArray[np.float64], 
        cluster_centers: npt.NDArray[np.float64], 
        labels:npt.NDArray[np.int64], 
        ord: int = 2
        ) -> np.float64:
    
    """
    inertia(X, cluster_centers, labels, ord=2)
    
    Calculates the inertia score
    
    This calculates the sum of the distance between all points and their cluster centers across the different time steps. See note.

    Parameters
    -----------
    X : numpy array
        Input time series data. Should be a 3 dimensional array in TNF fromat.
    cluster_centers : numpy array
        If numpy array, it is expected to be a 3D in TNF format. Here, N is the number of clusters. 
        If 2-D array, then it is interpreted as a K x F array where K is the number of clusters, and F is the number of features. Suitable for fixed cluster centers clustering.
    labels : numpy array 
        It is expected to be a 2D array of shape (N, T). Where N is the number of entities and T is the number of time steps. The value of the ith row at the t-th column is the label (cluster index) entity i was assigned to at time t.
        If 1-D array, it is interpreted as an array of length N. Where N is the number of entities. In such case, the i-th element is the cluster the i-th entit was assigned to across all time steps. Suitable for fixed assignment clustering.
    ord : int, default : 2
        The distance metric to use. 1 is l1 distance, 2 is l2 distance etc.

    Returns
    --------
    float
        The intertia value.
        
    Notes
    ------
    The inertia is calculated as: 
    
    .. math::
        \sum_{t=1}^{T} \sum_{i=1}^{N} D(X_{ti}, Z_t) 
    Where 
    `T`, `N` are the number of time steps and entities respectively, 
    `D` is a distance function (or metric e.g :math:`L_1` distance, :math:`L_2` distance etc), 
    :math:`X_{ti} \in \mathbf{R}^f` is the feature vector of entity `i` at time `t`,
    `f` is the number of features, and 
    :math:`Z_t \in \mathbf{R}^f` is the cluster center :math:`X_{ti}` is assigned to at time `t`

    See Also
    --------
    max_dist : Calculates the maximum distance
    """

    # X, _ = get_inferred_data(X)

    if isinstance(cluster_centers, list):
        cluster_centers = np.array([df.values for df in cluster_centers])

    if isinstance(labels, pd.DataFrame):
        labels = labels.values

    cluster_centers, labels = broadcast_data(X.shape[0], cluster_centers=cluster_centers, labels=labels)

    running_sum = 0

    for t in range(X.shape[0]):
       for k in range(cluster_centers.shape[1]):
            is_assigned = labels[:, t] == k
            dist = np.linalg.norm(X[t, :, :] - cluster_centers[t, k, :].reshape(-1, X.shape[2]), ord=ord, axis=1)

            running_sum += np.sum(dist * is_assigned)

    return running_sum

def max_dist(
        X: npt.NDArray[np.float64], 
        cluster_centers: npt.NDArray[np.float64], 
        labels: npt.NDArray[np.int64], 
        ord: int = 2) -> np.float64:

    """
    Calculate the max_dist score
        
    This calculates the maximum of the distance between all points and their cluster centers across the different time steps. See note.

    Parameters
    -----------
    X : numpy array
        Input time series data. Should be a 3 dimensional array in TNF fromat.
    cluster_centers : numpy array
        If numpy array, it is expected to be a 3D in TNF format. Here, N is the number of clusters. 
        If 2-D array, then it is interpreted as a K x F array where K is the number of clusters, and F is the number of features. Suitable for fixed cluster centers clustering.
    labels : numpy array 
        It is expected to be a 2D array of shape (N, T). Where N is the number of entities and T is the number of time steps. The value of the ith row at the t-th column is the label (cluster index) entity i was assigned to at time t.
        If 1-D array, it is interpreted as an array of length N. Where N is the number of entities. In such case, the i-th element is the cluster the i-th entit was assigned to across all time steps. Suitable for fixed assignment clustering.
    ord : int, default : 2
        The distance metric to use. 1 is l1 distance, 2 is l2 distance etc

    Returns
    --------
    float
        The max distance value.
        
    Notes
    ------
    The max_dist is calculated as: 
    
    .. math::
        max(D(X_{ti}, Z_t)) 

    Where 
    `D` is a distance function (or metric e.g :math:`L_1` distance, :math:`L_2` distance etc), 
    :math:`X_{ti} \in \mathbf{R}^f` is the feature vector of entity `i` at time `t`,
    `f` is the number of features, and 
    :math:`Z_t \in \mathbf{R}^f` is the cluster center :math:`X_{ti}` is assigned to at time `t`

    See Also
    --------
    interia : Calculates the inertia score
    """

    # X, _ = get_inferred_data(X)

    if isinstance(cluster_centers, list):
        cluster_centers = np.array([df.values for df in cluster_centers])

    if isinstance(labels, pd.DataFrame):
        labels = labels.values

    cluster_centers, labels = broadcast_data(X.shape[0], cluster_centers=cluster_centers, labels=labels)

    running_max = -np.inf

    for t in range(X.shape[0]):
       for k in range(cluster_centers.shape[1]):
            is_assigned = labels[:, t] == k
            dist = np.linalg.norm(X[t, :, :] - cluster_centers[t, k, :].reshape(-1, X.shape[2]), ord=ord, axis=1)

            max_d = np.max(dist * is_assigned)

            if max_d > running_max:
                running_max = max_d

    return running_max


from __future__ import annotations
from typing import TextIO, Tuple, List
import sys
import re
from time import time
import copy

import numpy as np
import numpy.typing as npt
from scipy.cluster.vq import kmeans, vq
import gurobipy as gp

from tscluster.interface import TSClusterInterface
from tscluster.base import TSCluster
from tscluster.opttscluster import optcluster
from tscluster.preprocessing.utils import infer_data

class OptTSCluster(TSCluster, TSClusterInterface):

    """
    Class for optimal time-series clustering. Throughout this doc and code, 'z' refers to cluster centers, while 'c' to label assignment.
    This creates an OptTSCluster object

    Parameters
    -----------
    n_clusters: int
        number of clusters
    scheme: {'z0c0', 'z0c1', 'z1c0', 'z1c1'}, default='z1c0'
        The scheme to use for tsclustering. Could be one of:
            - 'z0c0' means fixed center, fixed assignment
            - 'z0c1' means fixed center, changing assignment
            - 'z1c0' means changing center, fixed assignment
            - 'z1c1' means changing center, changing assignment
        Scheme needs to be a dynamic label assignment scheme (either 'z1c1' or 'z0c1') when using constrained cluster change (either with `n_allow_assignment_change` or `lagrangian_multiplier`)
    n_allow_assignment_change: int or None, default=None
        total number of label changes to allow
    use_sum_distance: bool, default=False
        Indicate if to use sum of distance to cluster as the objective. This is the sum of the distances between points in a time series
        and their centroids. 
    warm_start: bool, default=True
        Indicates if to use k-means to initialize the centroids (Z) and their assignments (C).
    use_MILP_centroid: bool, default=True
        If True, cluster_centers_ atrribute will be cluster centers obtained from MILP solution, else the average of the 
        datapoints per timestep
    random_state: int, default=None
        Set the random seed used when initializing with k-means or when initializing samples when using constraint generation.
    
    Attributes
    ----------
    cluster_centers_
    fitted_data_shape_
    labels_
    label_dict_
    n_changes_
    
    """

    def __init__(
            self, 
            n_clusters: int, 
            scheme: str = 'z1c0', 
            *,
            n_allow_assignment_change: None|int = None,  
            use_sum_distance: bool = False,
            warm_start: bool = True, 
            use_MILP_centroid: bool = True,
            random_state: None|int = None
            ) -> None:

        self.k = n_clusters
        self.scheme = scheme.lower()
        self.n_allow_assignment_change = n_allow_assignment_change
        self.is_Z_positive = True
        self.is_tight_constraints = True 
        self.lagrangian_multiplier = 0.0
        self.use_sum_distance = use_sum_distance
        self.warm_start = warm_start
        self.normalise_assignment_penalty = True
        self.strictly_n_allow_assignment_change = False
        self.use_MILP_centroid = use_MILP_centroid
        self.use_full_constraints = True
        self.IFrac = 0.2
        self.top_n_percentile = 0.0 
        self.max_iter = 10
        self.random_state = random_state
        self.add_constraint_per_cluster = True
        self.init_with_prev = True 

        self._use_closest_center_for_all = False

                # inital random X
        self.rand_gen = np.random.RandomState(self.random_state)

        if self.use_sum_distance:
            self.use_full_constraints = True 

        self.solver_schemes = {'z0c0', 'z1c0', 'z0c1', 'z1c1'}

        if self.scheme not in self.solver_schemes:
            raise ValueError(f"Invalid value for scheme. Expected any of {self.solver_schemes}, but got '{self.scheme}'")

        if re.search("c0", self.scheme, flags=re.IGNORECASE):
            self.n_allow_assignment_change = 0

        self.constant_assigment_constraint_violation_scheme = 'allow_v_violations_in_total'

        self.x_min = 0

    def fit(
            self, 
            X: npt.NDArray[np.float64], 
            label_dict: dict|None = None, 
            verbose: bool = True, 
            print_to: TextIO = sys.stdout,
            **kwargs
            ) -> "OptTSCluster": 

        """
        Method for fitting the model by solving the MILP model.
        
        Parameters
        -----------
        X : numpy array
            Input time series data. Should be a 3 dimensional array in TNF fromat.
        label_dict : dict, default=None
            A dictionary of the labels of X. Keys should be 'T', 'N', and 'F' (which are the number of time steps, entities, and features respectively). Value of each key is a list such that the value of key:
                - 'T' is a list of names/labels of each time step used as index of each dataframe during fit. Default is range(0, T). Where T is the number of time steps in the fitted data
                - 'N' is a list of names/labels of each entity used as index of the dataframe. Default is range(0, N). Where N is the number of entities/observations in the fitted data 
                - 'F' is a list of names/labels of each feature used as column of each dataframe. Default is range(0, F). Where F is the number of features in the fitted data 
            data_loader function from tscluster.preprocessing.utils can help in getting label_dict of a data. 
        verbose : bool, default=True
            If True, some model training information will be printed out. Set to False to surpress printouts
        print_to : TextIO, default=sys.stdout
            An object with a write method to write model's printout information during training. Default is standard output.

        Returns
        --------
        self : 
            The fitted OptTSCluster object.
        """

        self._label_dict_ = label_dict

        epsilon = 1e-4
        verbose_flush = True 

        self.verbose = verbose
        self.T_, self.N_, self.F_ = X.shape
        self.X_ = copy.deepcopy(X)

        X = copy.deepcopy(X)

        n = len(X)

        if np.min(X) < 0 and self.is_Z_positive:
            self.x_min = np.min(X) - epsilon

        X -= self.x_min

        X_2d = X.reshape(-1, X.shape[2])

        x_std = np.std(X_2d, axis=0)

        z_min = np.min(X_2d, axis=0) - 0.5*epsilon # to ensure Z is not negative
        z_max = np.max(X_2d, axis=0) + x_std

        e_max = np.abs(z_max - z_min)
        E_max = np.linalg.norm(e_max, ord=1)

        if not self.is_Z_positive:
            z_min = -gp.GRB.INFINITY
            z_max = gp.GRB.INFINITY

        if not self.is_tight_constraints:
            z_min = 0
            if not self.is_Z_positive:
                z_min = -gp.GRB.INFINITY

            z_max = gp.GRB.INFINITY
            e_max = gp.GRB.INFINITY

        if self.warm_start:
            init_Zs = []
            init_Cs = []

            if self.verbose:    
                print("Warm starting...", file=print_to, flush=verbose_flush)
                t0 = time()

            for t in range(X.shape[0]):
                # warm start clusters
                if len(np.unique(X[t, :, :])) == 1:
                    init_Zs.append(np.array([np.unique(X[t, :, :])[0]]*(self.k*X.shape[2])).reshape(self.k, X.shape[2]))

                    c_ = np.zeros((X.shape[1], self.k))
                    ran_c = self.rand_gen.choice(np.arange(self.k), X.shape[1])

                    c_[np.arange(len(ran_c)), ran_c] = 1
                    init_Cs.append(c_)

                    continue

                km_Z, _ = kmeans(X[t, :, :], self.k)
                km_labels, _ = vq(X[t, :, :], km_Z)

                # km_Z = km.cluster_centers_
                km_Z_to_Z_mapper = km_Z[:, 0].argsort()

                Z = km_Z[km_Z_to_Z_mapper, :]

                C = np.zeros((X.shape[1], self.k))

                km_Z_to_Z_mapper_key_val_pairs = np.array(list(enumerate(km_Z_to_Z_mapper)))

                for i in range(C.shape[0]):
                    j = km_Z_to_Z_mapper_key_val_pairs[:, 0][km_Z_to_Z_mapper_key_val_pairs[:, 1] == km_labels[i]] 
                    C[i, j] = 1      

                init_Zs.append(Z)
                init_Cs.append(C)  

            if self.verbose:    
                print(f"Done with warm start after {np.round(time() - t0, 2)}secs", file=print_to, flush=verbose_flush)        
                print(file=print_to, flush=verbose_flush)

            init_Zs = np.array(init_Zs)
            init_Cs = np.array(init_Cs)

            if re.search("z0", self.scheme, flags=re.IGNORECASE):
                init_Zs = np.mean(init_Zs, axis=0)
                init_Cs = OptTSCluster.assign_cluster(X, init_Zs)
            #     if self.scheme in ['z0c0', 'z0c0_1']:
            #         init_Cs = init_Cs[0, :, :]

            # elif re.search("z1c0", self.scheme, flags=re.IGNORECASE):
            #     init_Cs = OptTSCluster.assign_cluster(X, init_Zs[0, :, :])
            #     if self.scheme in ['z1c0', 'z1c0_1']:
            #         init_Cs = init_Cs[0, :, :]
        
        else:
            init_Zs = None
            init_Cs = None  

        # initialize X
        if self.use_full_constraints:
            I = X 
            I_idx = list(range(X.shape[1]))
        else:
            # if self.scheme in ['z1c1', 'z1c1_1']:
            #     n = X.shape[1]
            #     I = []
            #     I_idx_t = []
            #     for t in range(X.shape[0]):
            #         I_idx = list(self.rand_gen.choice(list(range(n)), size=np.int64(np.round(self.IFrac*n)), replace=False))
            #         I_idx_t.append(I_idx)
            #         I.append(X[t, I_idx, :])
            #     I = np.array(I)

            # elif self.scheme == 'z0c1_2':
            #     n = X.shape[0]*X.shape[1]

            #     I_idx = list(self.rand_gen.randint(0, n, np.int64(np.round(self.IFrac*n))))

            #     I = X.reshape(n, X.shape[2])[I_idx, :]

            # else:
            n = X.shape[1]
            I_idx = list(self.rand_gen.choice(list(range(n)), size=np.int64(np.round(self.IFrac*n)), replace=False))
            I = X[:, I_idx, :]
        
        E_star = np.array([0])
        E_hat = np.array([np.Inf])
        count = 1
        
        epsilon = epsilon 
        
        Es_star = []
        Zs_ans = []
        Cs_hats = []
        I_adds = []
        zis = []

        solver_times = []
        percent_data_used = []

        t0 = time()

        has_any_new_constraint = True

        if self.use_full_constraints:
            self.max_iter = 1

        while ((E_hat > E_star+epsilon).sum() > 0 or has_any_new_constraint) and count < self.max_iter+1:

            self.z_fixed = False
            if re.search("z0", self.scheme, flags=re.IGNORECASE):
                self.z_fixed = True

            self.c_fixed = False
            if re.search("c0", self.scheme, flags=re.IGNORECASE):
                self.c_fixed = True        

            _get_model_size = kwargs.get('_get_model_size', False)
                
            solver_time_0 = time()
            res = OptTSCluster.solve_ts_MILP(I, self.k, 
                                                                    n_allow_assignment_change=self.n_allow_assignment_change,
                                                                    z_fixed=self.z_fixed,
                                                                    _get_model_size=_get_model_size,
                            constant_assigment_constraint_violation_scheme=self.constant_assigment_constraint_violation_scheme,
                                                                    init_Z=init_Zs,
                                                                    init_C=init_Cs,
                                                strictly_n_allow_assignment_change=self.strictly_n_allow_assignment_change,
                                                z_max=z_max,
                                                z_min=z_min,
                                                e_max=e_max,
                                                E_max=E_max,
                                                is_Z_positive=self.is_Z_positive,
                                                is_tight_constraints=self.is_tight_constraints,
                                                lagrangian_multiplier=self.lagrangian_multiplier,
                                                normalise_assignment_penalty=self.normalise_assignment_penalty,
                                                use_sum_distance=self.use_sum_distance
                                                                    )
            
            if _get_model_size:
                return res 
            
            E_star, Z_ans, C_ans = res
            
            solver_times.append(np.round(time() - solver_time_0, 2))

            C_hat = OptTSCluster.assign_cluster(X, Z_ans) # C_hat contains assignment for all datapoints, while
                                                        # C_ans contains only for those used for solving the MILP.

            if not self._use_closest_center_for_all: #and self.scheme in ['z1c0']:
                C_hat[:, I_idx, :] = C_ans

            return_E_hat_per_t = False
            # if self.scheme in ['z1c1', 'z1c1_1']:
            #     return_E_hat_per_t = True  

            percent_data_used.append(np.round(100*I.size/X.size, 2))

            # if self.verbose:
            #     print(file=print_to, flush=verbose_flush)
            #     print(f"{percent_data_used[-1]}% of data used", file=print_to, flush=verbose_flush)

            if self.use_sum_distance:
                E_hat, I_add, zi = E_star, [], []
            else:
                E_hat, I_add, zi = OptTSCluster.add_constraints(X, Z_ans, C_hat, 
                                                                        n_allow_assignment_change=self.n_allow_assignment_change,
                    constant_assigment_constraint_violation_scheme=self.constant_assigment_constraint_violation_scheme,
                                                                        per_cluster=self.add_constraint_per_cluster,
                                                                        top_n_percentile=self.top_n_percentile,
                                                                        return_E_hat_per_t=return_E_hat_per_t,
                                                                        verbose=self.verbose,
                                                                        print_to=print_to)
            # if self.scheme in ['z1c1', 'z1c1_1']:
            #     I = []
            #     for t in range(X.shape[0]):
            #         I_idx_t[t].extend(I_add[t])
            #         I.append(X[t, I_idx_t[t], :])
            #     I = np.array(I)
                
            # else:
            # print("checking for new constraints", file=print_to, flush=verbose_flush) # remeber to remove
            I_idx = set(I_idx) 

            I_idx_ = copy.deepcopy(I_idx)
            I_idx.update(I_add)

            if I_idx_ == I_idx:
                has_any_new_constraint = False  
        
            I_idx = list(I_idx)       

                # if self.scheme == 'z0c1_2':
                #     Xt = X.reshape(X.shape[0]*X.shape[1], X.shape[2])
                #     I = Xt[I_idx, :]
                # else:
            I = X[:, I_idx, :]

            if self.verbose:    
                print(f"Obj val: {E_star}", file=print_to, flush=verbose_flush)
                print(file=print_to, flush=verbose_flush)
        
                count += 1

            if self.init_with_prev:
                init_Zs = Z_ans 
                init_Cs = C_hat  
                
            Es_star.append(E_star)
            Zs_ans.append(Z_ans + self.x_min)
            Cs_hats.append(C_hat)
            I_adds.append(I_add)
            zis.append(np.array(zi) + self.x_min)

        self.Es_star_ = Es_star
        self.Zs_ans_ = Zs_ans
        self.Cs_hats_ = Cs_hats
        self.I_adds_ = I_adds
        self.zis_ = zis

        self.C_ans_ = C_ans

        self.solver_times_ = solver_times
        self.n_iter_ = count - 1
        self.percent_data_used_ = percent_data_used 
        self.E_hat_ = E_hat 

        if self.warm_start:
            self.warm_start_Z_ = Z + self.x_min
            self.warm_start_labels_ = km_labels

        if self.verbose:
            print(f"Total time is {np.round(time() - t0, 2)}secs", file=print_to, flush=verbose_flush)  
            print(file=print_to, flush=verbose_flush)
        
        self.Zs_ans_avg = None 

        if not self.use_MILP_centroid:
            labels = self.labels_

            self.Zs_ans_avg = np.zeros((self.T_, self.k, self.F_))

            for t in range(self.T_):
                for j in range(self.k):
                    _idx = labels[:, t] == j
                    self.Zs_ans_avg[t, j, :] = np.mean(self.X_[t, _idx, :], axis=0)
    
        return self 
        
    @property
    def cluster_centers_(self):
        if self.use_MILP_centroid:
            return self.Zs_ans_[-1]
        else:
            return self.Zs_ans_avg

    @property
    def labels_(self):
        c = np.argmax(self.Cs_hats_[-1], axis=-1).T

        if self.c_fixed:
            return c[:, 0]
        return c

    @property
    def fitted_data_shape_(self) -> Tuple[int, int, int]:
        """
        returns a tuple of the shape of the fitted data in TNF format. E.g (T, N, F) where T, N, and F are the number of timesteps,
        observations, and features respectively. 
        """
        return self.T_, self.N_, self.F_

    def get_model_size(self,
            X: npt.NDArray[np.float64]
            ) -> Tuple:
        
        """
        Method to return the size of the model as a tuple of (v, c). Wehre v is the number of variables, and c is the number of constraints

        Parameters
        ---------
        X : numpy array
            Input time series data. Should be a 3 dimensional array in TNF fromat.        

        Returns
        -------
        number of variable
            The number of variables in the model
        number of constraints
            The number of constraints
        """
        return self.fit(X, _get_model_size=True, verbose=False)

    @staticmethod
    def solve_ts_MILP(
        I: npt.NDArray[np.float64] | npt.NDArray[np.int64], 
        k: int, 
        init_Z: npt.NDArray[np.float64] | npt.NDArray[np.int64] | None = None, 
        init_C: npt.NDArray[np.float64] | npt.NDArray[np.int64] | None = None, 
        _get_model_size: bool = False,
        **kwargs
        ) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:

        # MILP
        m = gp.Model()

        z_fixed = False
        if 'z_fixed' in kwargs:
            z_fixed = kwargs['z_fixed']

        v = kwargs['n_allow_assignment_change']
        lambda_val = kwargs['lagrangian_multiplier']
        normalise_assignment_penalty = kwargs['normalise_assignment_penalty']

        use_sum_distance = kwargs['use_sum_distance']

        T = I.shape[0]
        N = I.shape[1]
        F = I.shape[2]

        z_min = kwargs['z_min']
        z_max = kwargs['z_max']
        E_max = kwargs['E_max']
        E_max_ = kwargs['E_max']
        e_max = kwargs['e_max']

        if type(e_max) == type(gp.GRB.INFINITY): # to avoid error from 2D array e_max
            if e_max == gp.GRB.INFINITY:
                E_max = gp.GRB.INFINITY

        if kwargs['is_tight_constraints']:
            if kwargs['is_Z_positive']:
                z_min = np.vstack([z_min for _ in range(k)])
                z_max = np.vstack([z_max for _ in range(k)])

            e_max = np.vstack(
                [np.vstack(
                    [np.vstack([e_max for _ in range(k)]
                            ) for _ in range(N)
                    ]
                        ) for _ in range(T)
                ]
            )

        # creating Vars
        if use_sum_distance:
            E = m.addVars(T, N, name='E', lb=0, ub=E_max, vtype=gp.GRB.CONTINUOUS)
        else:
            E = m.addVar(name='E', lb=0, ub=E_max, vtype=gp.GRB.CONTINUOUS)

        if z_fixed:
            Z_vars = m.addVars(k, F, lb=z_min, ub=z_max, vtype=gp.GRB.CONTINUOUS)
            if init_Z is not None:
                for zj in range(k):
                    for zf in range(F):
                        Z_vars[zj, zf].start = init_Z[zj, zf]

        else:

            if kwargs['is_Z_positive'] and kwargs['is_tight_constraints']:
                z_min = np.vstack([z_min for _ in range(T)])
                z_max = np.vstack([z_max for _ in range(T)])

            Z_vars = m.addVars(T, k, F, lb=z_min, ub=z_max, vtype=gp.GRB.CONTINUOUS)
            if init_Z is not None:
                for zt in range(T):
                    for zj in range(k):
                        for zf in range(F):
                            Z_vars[zt, zj, zf].start = init_Z[zt, zj, zf]
                
        C_vars = m.addVars(T, N, k, vtype=gp.GRB.BINARY)
        if init_C is not None:
            for ct in range(T):
                for ci in range(N):
                    for cj in range(k):
                        C_vars[ct, ci, cj].start = init_C[ct, ci, cj]

        if v is not None or lambda_val > 0:
            y_vars = m.addVars(T-1, N, vtype=gp.GRB.BINARY)
                
        e = m.addVars(T, N, k, F, lb=0, ub=e_max, vtype=gp.GRB.CONTINUOUS) # to track absolute difference

        if z_fixed:
            # creating Contraints
            for t in range(T):
                for i in range(N):
                    for j in range(k):
                        for f in range(F):
            #                 e[i, j, f] = m.addVar(name='e_'+str(i)+'_'+str(j)+'_'+str(f))
                        
                            m.addConstr(e[t, i, j, f] >= I[t, i, f] - Z_vars[j, f])
                            m.addConstr(e[t, i, j, f] >= -(I[t, i, f] - Z_vars[j, f]))
        else:
            # creating Contraints
            for t in range(T):
                for i in range(N):
                    for j in range(k):
                        for f in range(F):
            #                 e[i, j, f] = m.addVar(name='e_'+str(i)+'_'+str(j)+'_'+str(f))
                        
                            m.addConstr(e[t, i, j, f] >= I[t, i, f] - Z_vars[t, j, f])
                            m.addConstr(e[t, i, j, f] >= -(I[t, i, f] - Z_vars[t, j, f]))
                    
        for t in range(T):
            for i in range(N):
                if use_sum_distance:
                    for j in range(k):
                        m.addConstr((C_vars[t, i, j] == 1) >> (E[t, i] >= gp.quicksum(e[t, i, j, f] for f in range(F))))

                else:
                    for j in range(k):
                        m.addConstr((C_vars[t, i, j] == 1) >> (E >= gp.quicksum(e[t, i, j, f] for f in range(F))))

                m.addConstr(gp.quicksum(C_vars[t, i, j] for j in range(k)) == 1)

                if t != T-1 and (v is not None or lambda_val > 0):
                    m.addConstr((y_vars[t, i] == 1) >> (gp.quicksum((jj+1)*C_vars[t, i, jj] for jj in range(k)) == 
                                    gp.quicksum((jj+1)*C_vars[t+1, i, jj] for jj in range(k))))

        if kwargs['constant_assigment_constraint_violation_scheme'] == 'allow_v_violations_in_total' and v is not None and lambda_val == 0:
            if kwargs['strictly_n_allow_assignment_change']:
                m.addConstr(gp.quicksum(y_vars[tt, ii] for tt in range(T-1) for ii in range(N)) == (T-1)*N-v)

            else:
                m.addConstr(gp.quicksum(y_vars[tt, ii] for tt in range(T-1) for ii in range(N)) >= (T-1)*N-v)
                

        elif kwargs['constant_assigment_constraint_violation_scheme'] == 'allow_v_violations_per_i' and v is not None and lambda_val == 0:
            for i in range(N):
                if kwargs['strictly_n_allow_assignment_change']:
                    m.addConstr(gp.quicksum(y_vars[tt, i] for tt in range(T-1)) == T-1-v)
                
                else:
                    m.addConstr(gp.quicksum(y_vars[tt, i] for tt in range(T-1)) >= T-1-v)

        # eps = 1e-3
        # if z_fixed:
        #         for j in range(k - 1):
        #             m.addConstr(Z_vars[j, 0] + eps <= Z_vars[j+1, 0])
        # else:

        #     for j in range(k - 1):
        #         m.addConstr(Z_vars[0, j, 0] + eps <= Z_vars[0, j+1, 0])
        
        if normalise_assignment_penalty:
            normaliser = E_max_ / ((T-1)*N)
        else:
            normaliser = 1

        if lambda_val > 0:
            if use_sum_distance:
                m.setObjective(gp.quicksum(E[tt, ii] for tt in range(T) for ii in range(N)) + 
                               lambda_val * normaliser * (1- gp.quicksum(y_vars[tt, ii] for tt in range(T-1) for ii in range(N))), 
                               gp.GRB.MINIMIZE)
            else:
                m.setObjective(E + lambda_val * normaliser * (1- gp.quicksum(y_vars[tt, ii] for tt in range(T-1) for ii in range(N))), 
                               gp.GRB.MINIMIZE)
        
        else:
            if use_sum_distance:
                m.setObjective(gp.quicksum(E[tt, ii] for tt in range(T) for ii in range(N)), gp.GRB.MINIMIZE)
            else:
                m.setObjective(E, gp.GRB.MINIMIZE)

        m.setParam('OutputFlag', 0)

        if _get_model_size:
            m.update()

            num_vars = m.NumVars
            num_constraints = m.NumConstrs

            return num_vars, num_constraints
        
        m.optimize() 
        
        if z_fixed:
            Z_ans = np.zeros((k, F))
        else:
            Z_ans = np.zeros((T, k, F))
        
        C_ans = np.zeros((T, N, k))

        for ji in range(k):
            for jj in range(F):
                if z_fixed:
                    Z_ans[ji, jj] = Z_vars[ji, jj].X
                else:
                    for jt in range(T):
                        Z_ans[jt, ji, jj] = Z_vars[jt, ji, jj].X

            for t in range(T):
                for ii in range(N):
                    C_ans[t, ii, ji] = C_vars[t, ii, ji].X

        if use_sum_distance:
            E_x = E.sum().getValue()      
        else: 
            E_x =E.X

        return np.array([E_x]), Z_ans, C_ans

    @staticmethod
    def assign_cluster(
        X: npt.NDArray[np.float64] | npt.NDArray[np.int64], 
        Z: npt.NDArray[np.float64], 
        # _use_closest_center_for_all=True, 
        **kwargs
        ) -> npt.NDArray[np.float64] | npt.NDArray[np.int64]:
        # Z_cluster = []
        C_hat = []

        T = X.shape[0]
        N = X.shape[1]
        F = X.shape[2]
        k = Z.shape[0]

        for t in range(T):
            if len(Z.shape) == 3:
                C_hat.append(optcluster.MILPClustering.assign_cluster(X[t], Z[t], **kwargs))
            else:
                C_hat.append(optcluster.MILPClustering.assign_cluster(X[t], Z, **kwargs))
            
        return np.array(C_hat)
    

    @staticmethod
    def add_constraints(
        X: npt.NDArray[np.float64] | npt.NDArray[np.int64], 
        Z: npt.NDArray[np.float64], 
        C_hat: npt.NDArray[np.float64] | npt.NDArray[np.int64], 
        per_cluster: bool = False, 
        top_n_percentile: float = 0.0, 
        return_E_hat_per_t: bool = False, 
        **kwargs
        ) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:

        E_hat_t = []

        if return_E_hat_per_t:
            I_add_t = []
        else:
            I_add_t = set()

        zi_t = []

        T = X.shape[0]
        N = X.shape[1]
        F = X.shape[2]
        k = Z.shape[0]

        for t in range(T):
            if len(Z.shape) == 3:
                E_hat, I_add, zi = optcluster.MILPClustering.add_constraints(X[t], Z[t], C_hat[t], 
                                                                 per_cluster=per_cluster, top_n_percentile=top_n_percentile)
            else:
                E_hat, I_add, zi = optcluster.MILPClustering.add_constraints(X[t], Z, C_hat[t], 
                                                                 per_cluster=per_cluster, top_n_percentile=top_n_percentile)
            
            E_hat_t.append(E_hat)

            if return_E_hat_per_t:
                I_add_t.append(I_add)
            else:
                I_add_t.update(I_add)

            zi_t.append(zi)
        
        if kwargs['n_allow_assignment_change'] is not None and kwargs['constant_assigment_constraint_violation_scheme'] is not None:

            I_add_t.update(OptTSCluster.constant_assigment_constraint_violators(C_hat, 
                                                                    v=kwargs['n_allow_assignment_change'],
                                                                    scheme=kwargs['constant_assigment_constraint_violation_scheme'],
                                                                    verbose=kwargs['verbose'],
                                                                    print_to=kwargs['print_to']
                                                                    )
            )
            
        if return_E_hat_per_t:
            E_hat_t = np.array(E_hat_t)
        else:
            E_hat_t = np.array([max(E_hat_t)])
            I_add_t = list(I_add_t)

        return E_hat_t, I_add_t, zi_t
    
    # @staticmethod
    # def add_constraints_Z_change_C_fixed(X, Z, C_hat, n_allow_assignment_change, constant_assigment_constraint_violation_scheme,
    #                                         per_cluster=False, top_n_percentile=False, return_E_hat_per_t=False, **kwargs):
    #     return OptTSCluster.add_constraints(X, Z, C_hat, n_allow_assignment_change=n_allow_assignment_change,
    #                                         constant_assigment_constraint_violation_scheme=constant_assigment_constraint_violation_scheme,
    #                                         per_cluster=False, top_n_percentile=False, return_E_hat_per_t=False, **kwargs)
        

    @staticmethod
    def constant_assigment_constraint_violators(
        C_hat: npt.NDArray[np.float64] | npt.NDArray[np.int64], 
        v: None|int, 
        scheme: str, 
        **kwargs
        ) -> List[int]:

        I_add = []
        # I_add_all = []

        verbose = kwargs['verbose']

        if scheme == "allow_v_violations_in_total":
            n_violations = 0
            for i in range(C_hat.shape[1]):
                y = np.argmax(C_hat[:, i, :], axis=1)
                if (y[:-1] != y[1:]).sum() > 0:
                    # if verbose:
                    #     print(f"i that changed: {i}", file=kwargs['print_to'])
                    I_add.append(i)
                    # I_add_all.append(i)
                    n_violations += (y[:-1] != y[1:]).sum()

            # if n_violations <= v:
            #     I_add = []
            
        elif scheme == 'allow_v_violations_per_i':
            for i in range(C_hat.shape[1]):
                y = np.argmax(C_hat[:, i, :], axis=1)
                if (y[:-1] != y[1:]).sum() > v:
                    # if verbose:
                    #     print(f"i that changed: {i}", file=kwargs['print_to'])
                    I_add.append(i)
                    # I_add_all.append(i)

        return I_add#, I_add_all

if __name__ == "__main__":
    pass


from __future__ import annotations
from typing import List, Any, Tuple


import numpy as np
import numpy.typing as npt
from sklearn.cluster import KMeans

from tscluster.interface import TSClusterInterface
from tscluster.base import TSCluster
from tscluster.preprocessing.utils import tnf_to_ntf, infer_data

class TSGlobalKmeans(KMeans, TSCluster, TSClusterInterface):
    # def __init__(self, *args, **kwargs):
    #     self._labels_ = None
    #     self._cluster_centers_ = None
    #     super().__init__(*args, **kwargs)

    """
    Applies sklearn's K-Means clustering to the version of a data  reshaped from a 3D array of shape T x N x F to a 2D array of shape (TxN) x F

    Read more in the Sklearn's user guide. The follow parameters are from sklearn's K-Means constructor.

    Parameters
    ----------
    n_clusters : int, default=8
        The number of clusters to form as well as the number of centroids to generate.

        For an example of how to choose an optimal value for n_clusters refer to
        sphx_glr_auto_examples_cluster_plot_kmeans_silhouette_analysis.py.

    init : {'k-means++', 'random'}, callable or array-like of shape (n_clusters, n_features), default='k-means++'
        Method for initialization:

    'k-means++' : selects initial cluster centroids using sampling based on an empirical probability distribution of the points' contribution to the overall inertia. This technique speeds up convergence. The algorithm implemented is "greedy k-means++". It differs from the vanilla k-means++ by making several trials at each sampling step and choosing the best centroid among them.

    'random': choose n_clusters observations (rows) at random from data for the initial centroids.

    If an array is passed, it should be of shape (n_clusters, n_features) and gives the initial centers.

    If a callable is passed, it should take arguments X, n_clusters and a random state and return an initialization.

        For an example of how to use the different init strategy, see the example
        entitled sphx_glr_auto_examples_cluster_plot_kmeans_digits.py.

    n_init : 'auto' or int, default='auto'
        Number of times the k-means algorithm is run with different centroid seeds. The final results is the best output of n_init consecutive runs in terms of inertia. Several runs are recommended for sparse
        high-dimensional problems (see kmeans_sparse_high_dim).

        When n_init='auto', the number of runs depends on the value of init: 10 if using init='random' or init is a callable; 1 if using init='k-means++' or init is an array-like.

    max_iter : int, default=300
        Maximum number of iterations of the k-means algorithm for a single run.

    tol : float, default=1e-4
        Relative tolerance with regards to Frobenius norm of the difference in the cluster centers of two consecutive iterations to declare convergence.

    verbose : int, default=0
        Verbosity mode.

    random_state : int, RandomState instance or None, default=None
        Determines random number generation for centroid initialization. Use an int to make the randomness deterministic.
        See Glossary <random_state>.

    copy_x : bool, default=True
        When pre-computing distances it is more numerically accurate to center the data first. If copy_x is True (default), then the original data is not modified. If False, the original data is modified, and put back before the function returns, but small numerical differences may be introduced by subtracting and then adding the data mean. Note that if the original data is not C-contiguous, a copy will be made even if copy_x is False. If the original data is sparse, but not in CSR format, a copy will be made even if copy_x is False.

    algorithm : {"lloyd", "elkan"}, default="lloyd"
        K-means algorithm to use. The classical EM-style algorithm is "lloyd". The "elkan" variation can be more efficient on some datasets with well-defined clusters, by using the triangle inequality. However it's more memory intensive due to the allocation of an extra array of shape (n_samples, n_clusters).
    
    Attributes
    ----------
    cluster_centers_
    fitted_data_shape_
    labels_
    label_dict_
    n_changes_
    """

    def fit(
        self, 
        X: npt.NDArray[np.float64],
        label_dict: dict|None = None, 
        **kwargs
        ) -> 'TSGlobalKmeans':
        """
        Method for fitting the model on the data.

        Parameters
        ----------
        X : numpy array
            Input time series data. Should be a 3 dimensional array in TNF fromat.
        label_dict : dict, default=None
            A dictionary of the labels of X. Keys should be 'T', 'N', and 'F' (which are the number of time steps, entities, and features respectively). Value of each key is a list such that the value of key:
                - 'T' is a list of names/labels of each time step used as index of each dataframe during fit. Default is range(0, T). Where T is the number of time steps in the fitted data
                - 'N' is a list of names/labels of each entity used as index of the dataframe. Default is range(0, N). Where N is the number of entities/observations in the fitted data 
                - 'F' is a list of names/labels of each feature used as column of each dataframe. Default is range(0, F). Where F is the number of features in the fitted data 
            data_loader function from tscluster.preprocessing.utils can help in getting label_dict of a data. 
        **kwargs keyword arguments to be passed to skleanr's kmeans fit method

        Returns
        -------
        self: 
            The fitted TSGlobalKmeans object. 
        """

        self._label_dict_ = label_dict

        self._labels_ = None
        self._cluster_centers_ = None

        self.Xt = tnf_to_ntf(X)

        self.N, self.T, self.F = self.Xt.shape

        self.Xt = np.vstack(self.Xt)

        super().fit(self.Xt, **kwargs) 

        return self

    @property
    def cluster_centers_(self) -> npt.NDArray[np.float64]: 
        return self._cluster_centers_
    
    @cluster_centers_.setter
    def cluster_centers_(self, new_value: npt.NDArray[np.float64]) -> None:
        self._cluster_centers_ = new_value

    @property
    def labels_(self) -> npt.NDArray[np.int64]:
        if self._labels_ is not None:
            return self._labels_.reshape(self.N, self.T)
        
        return self._labels_
    
    @labels_.setter
    def labels_(self, new_value: Any) -> None:
        self._labels_ = new_value

    @property
    def fitted_data_shape_(self) -> Tuple[int, int, int]:
        """
        returns a tuple of the shape of the fitted data in TNF format. E.g (T, N, F) where T, N, and F are the number of timesteps, observations, and features respectively. 
        """
        
        return self.T, self.N, self.F


from __future__ import annotations
from typing import List, Callable, Any, Tuple
import os 
from functools import wraps

import numpy as np 
import numpy.typing as npt
import pandas as pd

valid_data_load_types = {np.ndarray, pd.DataFrame, str, list}
valid_data_load_types_names = tuple[npt.NDArray, int](map(lambda x: x.__name__, valid_data_load_types))

default_data_loader_args = {
    "arr_format": "TNF", 
    "suffix_sep": "_", 
    "read_file_args": {},
    "file_reader": "infer", # can be one of ("infer", "np_load", "pd_read_csv", "pd_read_json", pd_read_excel)
    'use_suffix_as_label': False
    }

file_readers = {
    "np_load": np.load,
    "pd_read_csv": pd.read_csv,
    "pd_read_json": pd.read_json,
    "pd_read_excel": pd.read_excel
}

def tnf_to_ntf(X: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """
    Utility function to convert an array from Time x Number of observation x Feature format to Number of observation x Time x Feature format
    """

    T, N, F = X.shape 

    Xt = np.zeros(shape=(N, T, F))

    for t in range(T):
        Xt[:, t, :] = X[t, :, :]

    return Xt 

def ntf_to_tnf(X: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """
    Utility function to convert an array from Number of observation x Time x Feature format to Time x Number of observation x Feature format         
    """

    N, T, F = X.shape 

    Xt = np.zeros(shape=(T, N, F))

    for t in range(T):
        Xt[t, :, :] = X[:, t, :]

    return Xt 

def reshape_for_transform(X: npt.NDArray[np.float64], per_time: bool) -> Tuple[npt.NDArray, int]:
    """
    Reshape to appropriate shape for transformation. Assumes input shape is TNF
    """
    if per_time:
        n = X.shape[0]

    else:
        n = 1
        X = np.array([np.vstack(X)])

    return X, n 

def to_tnf(X: npt.NDArray[np.float64], arr_format: str, label_dict: dict) -> Tuple[npt.NDArray[np.float64], dict]:
    """
    Utility function to check the format of an array and converts it to TNF format. Raises ValueError if the array is not 3d.
    """
    if X.ndim != 3:
        raise ValueError(f"Invalid dimension of array. Expected array with 3 dimensions but got {X.ndim}")
    
    elif arr_format.upper() == 'NTF':
        label_dict['T'], label_dict['N'] = label_dict['N'], label_dict['T']
        
        return ntf_to_tnf(X), label_dict
    
    return X, label_dict

def broadcast_data(
        T: int,
        *,
        cluster_centers: npt.NDArray[np.float64]|None = None, 
        labels: npt.NDArray[np.int]|None = None, 
        ) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.int64]]:
    
    """function to make cluster_centers of shape T x N x F and labels of shape N x F """

    if cluster_centers is not None:
        if cluster_centers.ndim == 2:
            cluster_centers = np.array([cluster_centers for _ in range(T)])

    if labels is not None:
        if labels.ndim == 1:
            labels = np.array([labels for _ in range(T)]).T

    return cluster_centers, labels


def is_all_type(lst: List, data_type: type, type_checker: Callable[[Any, type], bool]=isinstance) -> bool:
    """
    Utility function to check if all elements of a list are of the same data type
    """
    return all(map(lambda x: type_checker(x, data_type), lst))

def get_lst_of_filenames(dir: str) -> List[str]:
    """
    Utility function to return a list of all filenames in a directory
    """
    return [f for f in os.listdir(dir) if os.path.isfile(os.path.join(dir, f))]

def get_default_header(file_reader: str, kwargs: dict) -> dict:
    """
    Utility function to set the default value header for pandas.read_csv or read_excel to None
    """
    if file_reader in ('pd_read_csv', 'pd_read_excel') and 'header' not in kwargs:
        kwargs['header'] = None

    return kwargs

def read_all_files(lst: List[str], file_reader: str, **kwargs) -> Tuple[npt.NDArray[np.float64], dict]:
    """
    Utility function to read all the files in a list
    """
    if file_reader == 'infer':

        file_extension = {filename.split('.')[-1].lower() for filename in lst}

        if 'npy' in file_extension or 'npz' in file_extension:
            df_lst = [pd.DataFrame(file_readers['np_load'](file_path, **kwargs)) for file_path in lst]

        elif 'json' in file_extension:
            df_lst = [file_readers['pd_read_json'](file_path, **kwargs).sort_index() for file_path in lst]


        elif 'xls' in file_extension or 'xlsx' in file_extension:
            kwargs = get_default_header('pd_read_excel', kwargs)
            df_lst = [file_readers['pd_read_excel'](file_path, **kwargs).sort_index() for file_path in lst]
        
        else: # assume csv
            kwargs = get_default_header('pd_read_csv', kwargs)
            df_lst = [file_readers['pd_read_csv'](file_path, **kwargs).sort_index() for file_path in lst]   
        
    elif file_reader == 'np_load': # if numpy
        df_lst = [pd.DataFrame(file_readers['np_load'](file_path, **kwargs)) for file_path in lst]
    
    else: # if pandas
        kwargs = get_default_header(file_reader, kwargs)
        df_lst = [file_readers[file_reader](file_path, **kwargs).sort_index() for file_path in lst]   

    df = pd.concat(df_lst, axis=0, sort=False)
    label_dict = {
        'T': list(range(len(lst))),
        'N': list(df_lst[0].index),
        'F': list(df.columns)
        }
    return df.values.reshape(len(lst), -1, df.shape[1]), label_dict 

def get_infer_data_wrapper_args(arg: str, kwargs: dict) -> Any:
    """
    Utility function to get the value of arg if arg is an argument for the decorator e.g. argument for data loader
    """
    try:
        arg_value = kwargs.pop(arg)
    except KeyError:
        arg_value = default_data_loader_args[arg] 

    return arg_value   

def infer_data(func: Callable) -> Callable:
    
    # Decorator to infer the data type of X, load it and return it in TNF format.
    
    @wraps(func)
    def args_selector(
            self: Any, 
            X: npt.NDArray[np.float64]|str|List, 
            *args, 
            **kwargs
            ) -> Any:
        
        # Decorator to infer the data type of X, load it and return it in TNF format.
    
        def data_loader(
                self: Any, 
                X: npt.NDArray[np.float64]|str|List, 
                arr_format: str, 
                suffix_sep: str, 
                file_reader: str,
                read_file_args: dict, 
                use_suffix_as_label: bool,
                *args, 
                **kwargs
                ) -> Any:
            
            """
            Parameters
            -----------
            X: ndarray, string or list. 
                Input time series data. If ndarray, should be a 3 dimensional array. If str and a file name, will use numpy to load file.
                If str and a directory name, will load all the files in the directory in ascending order of the suffix of the filenames.
                Use suffix_sep as a keyword argument to indicate the suffix separator. Default is "_". So, file_0.csv will be read first before file_1.csv and so on.
                Supported files in the directory are any file that can be read using any of np.load, pd.read_csv, pd.read_json, and pd.read_excel.
                If list, assumes the list is a list of files or filepaths. If file, each should be a numpy array or pandas DataFrame of data for the different time steps.
                If list of filepaths, data is read in the order in the list using any of np.load, pd.read_csv, pd.read_json, and pd.read_excel.

            arr_format: str, default 'TNF'
                format of the loaded data. 'TNF' means the data dimension is Time x Number of observations x Features
                'NTF' means the data dimension is Number OF  observations x Time x Features
            suffix_sep: str, default '_'
                separator separating the file number from the filename.
            file_reader: str, default 'infer'
                file loader to use. Can be any of np.load, pd.read_csv, pd.read_json, and pd.read_excel. If 'infer', decorator will attempt to infer the file type from the file name 
                and use the approproate loader.
            read_file_args: dict, default empty dictionary.
                parameters to be passed to the data loader.
            use_suffix_as_label: bool, default False
                If True, use the suffixes of the file names as labels in `label_dict`. If `arr_format` = 'TNF', the suffixes will be used as labels of timesteps, else if `arr_format` = 'NTF', they will be used as labels of entities.
                If False, a linear range of the number of files is used as labels (e.g.) range(n_files), where n_files is the number of files.
            *args: 
                any other positional arguments for fit method or function to be decorated.

            **kwargs:
                any keyword argument for fit method, or function to be decorated.
            """

            if isinstance(X, np.ndarray):
                X_arr = X 
                label_dict = {
                    'T': list(range(X_arr.shape[0])),
                    'N': list(range(X_arr.shape[1])),
                    'F': list(range(X_arr.shape[2]))
                }
            
            elif isinstance(X, list):

                if is_all_type(X, np.ndarray):
                    X_arr = np.array(X)
                    label_dict = {
                        'T': list(range(X_arr.shape[0])),
                        'N': list(range(X_arr.shape[1])),
                        'F': list(range(X_arr.shape[2]))
                    }

                elif is_all_type(X, pd.DataFrame):
                    df_lst = [df.sort_index() for df in X]
                    df = pd.concat(df_lst, axis=0, sort=False)
                    label_dict = {
                        'T': list(range(len(X))),
                        'N': list(df_lst[0].index),
                        'F': list(df.columns)
                        }
                    X_arr = df.values.reshape(len(X), -1, df.shape[1])
                    
                elif is_all_type(X, str):
                    X_arr, label_dict = read_all_files(X, file_reader, **read_file_args)

            elif isinstance(X, str):

                if os.path.isfile(X):
                    X_arr = file_readers['np_load'](X, **read_file_args)
                    label_dict = {
                        'T': list(range(X_arr.shape[0])),
                        'N': list(range(X_arr.shape[1])),
                        'F': list(range(X_arr.shape[2]))
                    }

                else:
                    file_names = get_lst_of_filenames(X)

                    file_list_sort_key = lambda filename: int("".join(filename.split(".")[0]).split(suffix_sep)[-1])

                    sorted_filenames = sorted(file_names, key=file_list_sort_key)

                    lst_of_filepaths = [os.path.join(X, f) for f in sorted_filenames]

                    X_arr, label_dict = read_all_files(lst_of_filepaths, file_reader, **read_file_args)

                    if use_suffix_as_label:
                        label_dict['T'] = sorted([str(file_list_sort_key(f)) for f in file_names])
            
            else:
                raise TypeError(f"Invalid type! Expected any of {valid_data_load_types_names}, but got '{type(X).__name__}'")
            
            X_arr, label_dict = to_tnf(X_arr, arr_format, label_dict) 

            try:
                self._label_dict_ = label_dict
            except AttributeError:
                self.append(label_dict)

            return func(self, X_arr, *args, **kwargs)
        
        arr_format = get_infer_data_wrapper_args('arr_format', kwargs)
        suffix_sep = get_infer_data_wrapper_args('suffix_sep', kwargs)
        file_reader = get_infer_data_wrapper_args('file_reader', kwargs)
        read_file_args = get_infer_data_wrapper_args('read_file_args', kwargs)
        use_suffix_as_label = get_infer_data_wrapper_args('use_suffix_as_label', kwargs)
            
        return data_loader(self, X, arr_format, suffix_sep, file_reader, read_file_args, use_suffix_as_label, *args, **kwargs)
    
    return args_selector

@infer_data
def _get_inferred_data(label_dict: dict, X: npt.NDArray[np.float64]|str|List, **kwargs) -> np.float64:
    """
    function to replace self argument
    """
    return X

def load_data(
        X: npt.NDArray[np.float64]|str|List, 
        *,
        arr_format: str = 'TNF',
        suffix_sep: str = '_',
        file_reader: str = 'infer',
        read_file_args: dict|None = None,
        use_suffix_as_label: bool = False,
        output_arr_format: str = 'TNF'
        ) -> Tuple[np.float64, dict]:
    """
    function to load data

    Parameters
    ----------
    X : numpy array, string or list
        Input time series data. If ndarray, should be a 3 dimensional array, use `arr_format` to specify its format. If str and a file name, will use numpy to load file.
        If str and a directory name, will load all the files in the directory in ascending order of the suffix of the filenames.
        Use suffix_sep as a keyword argument to indicate the suffix separator. Default is "_". So, file_0.csv will be read first before file_1.csv and so on.
        Supported files in the directory are any file that can be read using any of np.load, pd.read_csv, pd.read_json, and pd.read_excel.
        If list, assumes the list is a list of files or filepaths. If file, each should be a numpy array or pandas DataFrame of data for the different time steps.
        If list of filepaths, data is read in the order in the list using any of np.load, pd.read_csv, pd.read_json, and pd.read_excel.
    arr_format : str, default='TNF'
        format of the input data. 'TNF' means the data dimension is Time x Number of observations x Features
        'NTF' means the data dimension is Number OF  observations x Time x Features
    suffix_sep : str, default='_'
        separator separating the suffix from the filename. The suffixes should be numbers and may not neccessarily need to start from 1 or have an interval of 1. So long the suffixes can be sorted and there is a consistent suffix separator, the directory can be parsed by `load_data` function.
    file_reader : str, default='infer'
        file loader to use. Can be any of np.load, pd.read_csv, pd.read_json, and pd.read_excel. If 'infer', decorator will attempt to infer the file type from the file name 
        and use the approproate loader.
    read_file_args : dict, default=None.
        parameters to be passed to the data loader. Keys of the dictionary should be parameter names as keys in str, values should be the values of the parameter keys.
    use_suffix_as_label: bool, default=False
        If True, use the suffixes of the file names as labels in `label_dict`. If `arr_format` = 'TNF', the suffixes will be used as labels of timesteps, else if `arr_format` = 'NTF', they will be used as labels of entities.
        If False, a linear range of the number of files is used as labels (e.g.) range(n_files), where n_files is the number of files.
    output_arr_format : str, default='TNF'
        The format of the output array. Can be any of {'TNF', 'NTF'}.

    Returns
    --------
    np.array 
        a numpy array of the data in 'TNF' or 'NTF' format (depending on the value of `output_arr_format`)
    dict
        a dictionary whose keys are 'T', 'N', and 'F', and whose values are lists of the labels of each key. 
    """

    label_dict = []

    if read_file_args is None:
        read_file_args = {}

    X_arr = _get_inferred_data(
        label_dict, 
        X, 
        arr_format=arr_format, 
        suffix_sep=suffix_sep,
        file_reader=file_reader,
        read_file_args=read_file_args,
        use_suffix_as_label=use_suffix_as_label
        )

    if output_arr_format == 'NTF':
        X_arr = tnf_to_ntf(X_arr)

    return X_arr, label_dict[0]

def to_dfs(
        X: npt.NDArray[np.float64],
        label_dict: dict|None = None,
        arr_format: str = 'TNF',
        output_df_format: str = 'TNF'
        ):
    
    """
    Function to convert from (numpy, label_dict) to dataframes

    Parameters
    ----------
    X : numpy array
        The array to be converted to list of dataframes. Should be a 3D array.
    label_dict dict, default=None
        a dictionary whose keys are 'T', 'N', and 'F' (which are the number of time steps, entities, and features respectively). Value of each key is a list such that the value of key:
        - 'T' is a list of names/labels of each time step to be used as index of each dataframe. If None, range(0, T) is used. Where T is the number of time steps in the fitted data
        - 'N' is a list of names/labels of each entity to be used as index of the dataframe. If None, range(0, N) is used. Where N is the number of entities/observations in the fitted data 
        - 'F' is a list of names/labels of each feature to be used as column of each dataframe. If None, range(0, F) is used. Where F is the number of features in the fitted data 
        If label_dict is None, a linear range of the dimensions of the array is used.
    arr_format : str, default 'TNF'
        format of the array. 'TNF' means the data dimension is Time x Number of observations x Features
        'NTF' means the data dimension is Number OF  observations x Time x Features
    output_df_format : str, default='TNF'
        The format of the output dataframes. Can be any of {'TNF', 'NTF'}. If 'TNF', output is a list of T dataframes each of shape (N, F). If 'NTF', output is a list of N dataframes each of shape (T, F).

    Returns
    -------
    list[pd.DataFrame]
        A list of T pandas DataFrames. Where T is the number of time steps. The t-th dataframe in the list is a N x F dataframe of the values of the time series data of all entities at the t-th timestep.    
    """

    # arr_format will help us to interpret the array and convert it to a common format i.e TNF
    if arr_format == 'NTF':
        Xt = ntf_to_tnf(X)

        arr_lst = [Xt[t] for t in range(Xt.shape[0])]

    elif arr_format == 'TNF':
        arr_lst = [X[t] for t in range(X.shape[0])]

    else:
        raise ValueError(f"Expected arr_format to be any of {'TNF', 'NTF'} but got '{arr_format}'")

    dimensions = ('T', 'N', 'F')

    # get label_dict ready
    if label_dict is None:
        label_dict = {k: list(range(np.array(arr_lst).shape[i])) for i, k in enumerate(dimensions)}

    else:
        for i, k in enumerate(dimensions):
            if label_dict[k] is None:
                label_dict[k] = list(range(np.array(arr_lst).shape[i]))

    # output to required format
    if output_df_format == 'NTF':
        Xt = tnf_to_ntf(np.array(arr_lst))

        df_lst = [pd.DataFrame(Xt[i], index=label_dict['T'], columns=label_dict['F']) for i in range(Xt.shape[0])]

    elif output_df_format == 'TNF':
        df_lst = [pd.DataFrame(Xt, index=label_dict['N'], columns=label_dict['F']) for Xt in arr_lst]

    else:
        raise ValueError(f"Expected output_df_format to be any of {'TNF', 'NTF'} but got '{output_df_format}'")

    return df_lst


from __future__ import annotations
from typing import List, Tuple

import numpy as np
import numpy.typing as npt
import matplotlib
import mpl_toolkits
import matplotlib.pyplot as plt
import matplotlib.pylab as pl
from matplotlib.widgets import Slider

from tscluster.preprocessing.utils import broadcast_data

def _data_validator(
        X: npt.NDArray[np.float64]|None = None, 
        cluster_centers: npt.NDArray[np.float64]|None = None, 
        labels: npt.NDArray[np.float64]|None = None        
    ) -> None:

    """function to check the shapes of the data. Raises error if there are any inconsistenties in dimension of the data"""

    data = (X, cluster_centers, labels)
    valid_shapes = [{3}, {2, 3}, {2, 1}]
    names = ('X', 'cluster_centers', 'labels') 

    for i, j, k in zip(data, valid_shapes, names):
        if i is not None:
            if i.ndim not in j:
                raise TypeError(f"Invalid ndim. Expected {k}'s dimension to be any of {j} but got {i.ndim}")

    if cluster_centers is not None and labels is not None:
        if cluster_centers.shape[-2] != len(np.unique(labels)):
            raise ValueError(f"Number of clusters in cluster_centers and labels are not the same, they are {cluster_centers.shape[-2]} and {len(np.unique(labels))} respectively")
        elif cluster_centers.ndim == 3 and labels.ndim == 2 and cluster_centers.shape[0] != labels.shape[1]:
            raise ValueError(f"Number of timesteps in cluster_centers and labels are not the same, they are {cluster_centers.shape[0]} and {labels.shape[1]} respectively")

    if cluster_centers is not None and X is not None:
        if cluster_centers.shape[-1] != X.shape[-1]:
            raise ValueError(f"Number of features in cluster_centers and input data (X) are not the same, they are {cluster_centers.shape[-1]} and {X.shape[-1]} respectively")
        elif cluster_centers.ndim == 3 and cluster_centers.shape[0] != X.shape[0]:
            raise ValueError(f"Number of timesteps in cluster_centers and input data (X) are not the same, they are {cluster_centers.shape[0]} and {X.shape[0]} respectively")
     
    if labels is not None and X is not None:
        if labels.shape[0] != X.shape[1]:
            raise ValueError(f"Number of entities in labels and input data (X) are not the same, they are {labels.shape[0]} and {X.shape[1]} respectively")
        elif labels.ndim == 2 and labels.shape[1] != X.shape[0]:
            raise ValueError(f"Number of timesteps in labels and input data (X) are not the same, they are {labels.shape[1]} and {X.shape[0]} respectively")

def _get_shape(
        X: npt.NDArray[np.float64]|None = None, 
        cluster_centers: npt.NDArray[np.float64]|None = None, 
        labels: npt.NDArray[np.float64]|None = None 
    ) -> Tuple[Tuple[int, int, int], Tuple[int, int, int], Tuple[int, int]]:

    """
    Function to return the shape of the input data
    """

    if X is not None:
        X_shape = X.shape
    else:
        X_shape = (0, 0, 0)

    if cluster_centers is not None:
        if cluster_centers.ndim == 3:
            cc_shape = cluster_centers.shape 
        elif cluster_centers.ndim == 2:
            cc_shape = (0, *cluster_centers.shape)
    else:
        cc_shape = (0, 0, 0)

    if labels is not None:
        if labels.ndim == 2:
            l_shape = labels.shape 
        elif labels.ndim == 1:
            l_shape = (labels.shape[0], 0)
    else:
        l_shape = (0, 0)

    return X_shape, cc_shape, l_shape

def plot(
        *,
        X: npt.NDArray[np.float64]|None = None, 
        cluster_centers: npt.NDArray[np.float64]|None = None, 
        labels: npt.NDArray[np.float64]|None = None, 
        entity_idx: List[int]|None = None,
        entities_labels: List[str]|None = None,
        label_dict: dict|None = None,
        annot_fontsize: float|int = 10,
        show_all_entities: bool = True,
        figsize: Tuple[float, float] | None = None,
        shape_of_subplot: Tuple[int, int]|None = None, 
        xlabel: str|None = 'timesteps', 
        ylabel: str|None = 'value',
        cluster_labels: List[str]|None = None,
        title_list: List[str]|None = None,
        show_all_xticklabels: bool = True, 
        x_rotation: float|int = 45,
        show_X_marker: bool = False,
        show_cluster_center_marker: bool = False,
        ) -> Tuple[matplotlib.figure.Figure, List[matplotlib.axes.Axes]]:
    
    """
    Function to plot subplots of the time series data, cluster centers and label assignemnts. One subplot per feature. This is built on top of matplotlib.


    Parameters
    ----------
    X : numpy array, default=None
        The time series data in TNF format.
    cluster_centers : numpy array, default=None
        If numpy array, it is expected to be a 3D in TNF format. Here, N is the number of clusters. 
        If 2-D array, then it is interpreted as a K x F array where K is the number of clusters, and F is the number of features. Suitable for fixed cluster centers clustering.
    labels : numpy array, default=None
        It is expected to be a 2D array of shape (N, T) . Where N is the number of entities and T is the number of time steps. The value of the ith row at the t-th column is the label (cluster index) entity i was assigned to at time t.
        If 1-D array, it is interpreted as an array of length N. Where N is the number of entities. In such case, the i-th element is the cluster the i-th entity was assigned to across all time steps. Suitable for fixed assignment clustering.
    entity_idx : list, default=None 
        list of index of entities to display in the plot. If `show_all_entities` is True, `entity_idx` will be interpreted as the index of entities to be annonated.
    entities_labels : list, default=None
        list of labels for annotating the entities in `entity_idx`. If None, then labels of `entity_idx` in `label_dict` are used.
    label_dict dict, default=None
        a dictionary whose keys are 'T', 'N', and 'F' (which are the number of time steps, entities, and features respectively). Value of each key is a list such that the value of key:
        - 'T' is a list of names/labels of each time step to be used as index of each dataframe. If None, range(0, T) is used. Where T is the number of time steps in the fitted data
        - 'N' is a list of names/labels of each entity. If None, range(0, N) is used. Where N is the number of entities/observations in the fitted data 
        - 'F' is a list of names/labels of each feature to be used as column of each dataframe. If None, range(0, F) is used. Where F is the number of features in the fitted data 
        If label_dict is not None, it is used to label timestep, entities, and features in the plot.   
    annot_fontsize : float|int, default=10
        The font size to be used for annotating entities in `entity_idx`.
    show_all_entities : bool, default=True
        If True, displays all the entities in `X` in the plot. If False, only entities in `entity_idx` are plotted.
    figsize : tuple, default=None
        The size of the figure. Should be tuple of length 2, where the first value is width of the figure, while the second value is the height of the figure.
    shape_of_subplot : tuple, default=None
        The shape of the subplots. Should be tuple of length 2, where the first value is number of rows, while the second value is the of columns in the figure.
        If None, (F, 1) is used. Where F is the number of features in `X` or `cluster_centers`.
    xlabel : str, default='timesteps'
        The label for the x-axis of each subplots
    ylabel : str, default='timesteps'
        The label for the y-axis of each subplots
    cluster_labels : list, default=None
        The labels to use for the clusters. If None, list(range(K)) is used. Where `K` is the number of clusters in `cluster_centers` or `labels`.
    title_list : list, default=None
        The title of each subplot (or feature).
    show_all_xticklabels : bool, default=True
        If True, shows labels of all time steps in the plot. If False, some may be suppressed (depending on the size of the plot)
    x_rotation : float or int, default=45
        The angle (in degrees) to rotate the timestep labels 
    show_X_marker : bool, default=False
        If True, show markers in the time series plot of X.
    show_cluster_center_marker : bool, default=False
        If True, show markers in the time series plot of the cluster centers.

    Returns
    -------
    matplotlib.figure.Figure
        the matplotlib figure object
    list
        a list of the matplotlib axes, one axes per subplot. 
    """
    
    _data_validator(X=X, cluster_centers=cluster_centers, labels=labels)

    X_shape, cc_shape, l_shape = _get_shape(X=X, cluster_centers=cluster_centers, labels=labels)

    # determine T. Should be the longest in X, cluster_centers and labels. this is to allow for variable number of input variables ie X, cluster_centers and labels
    T = max(X_shape[0], cc_shape[0], l_shape[1])
    N = max(X_shape[1], l_shape[0])
    F = max(X_shape[2], cc_shape[2])
    K = cc_shape[1]
    
    if labels is not None:
        K = max(K, len(np.unique(labels)))

    K = max(K, 1)

    label_dict_init = {'T': T, 'N': N, 'F': F}

    if label_dict is None:
        label_dict = {}
    
    for key, val in label_dict_init.items():
            _ = label_dict.setdefault(key, list(range(val)))

    cmap = pl.cm.get_cmap('rainbow')

    # broadcast cluster centers and labels if need be. This is done at this point because we need to compute T before now
    cluster_centers, labels = broadcast_data(T, cluster_centers=cluster_centers, labels=labels)

    norm = plt.Normalize(vmin=0, vmax=K-1)
    
    if cluster_labels is None:
        cluster_labels = list(map(str, range(K)))

    # determine the number of feature

    fig = plt.figure(figsize=figsize)

    if shape_of_subplot is None:
        shape_of_subplot = (F, 1)

    if title_list is None:
        title_list = ['Feature ' + str(f+1) if isinstance(f, int) else 'Feature ' + f for f in label_dict['F']]

    axes = []

    for f in range(F):
        ax = fig.add_subplot(*shape_of_subplot, f+1)

        if X is not None:
            if show_all_entities:
                idx = np.arange(X.shape[1])
            else:
                idx = entity_idx

            marker = ''
            if show_X_marker:
                marker = '.'

            # plot all data for feature f
            plt.plot(range(X.shape[0]), X[:, idx, f], c='k', ls='--', alpha=0.5, marker=marker)

            if entity_idx is not None:
                for li, i in enumerate(entity_idx):
                    if entities_labels is None:
                        e_labels = label_dict['N'][i]
                    else:
                        e_labels = entities_labels[li]

                    annot_i = np.random.choice(np.arange(len(X[:, i, f])), 1)[0]
                    annot_xy = list(enumerate(X[:, i, f]))[annot_i]

                    plt.annotate(e_labels, xy=annot_xy, xytext=(annot_xy[0]+0.5, annot_xy[1]+0.5), fontsize=annot_fontsize,
                                arrowprops=dict(facecolor='green',shrink=0))
                    
            if labels is not None:
            # scatter plot for marker for label assignment of data points. 
                for i in idx:
                    c = labels[i] 
                    plt.scatter(range(X.shape[0]), X[:, i, f], color=cmap(norm(c)), s=10)

                # label legend for cluster centers to match that of label assignment
                if cluster_centers is None:
                    for j in range(K):
                        plt.plot([], [], color=cmap(norm(j)), label=cluster_labels[j])  

        # plot of cluster centers
        if cluster_centers is not None:

            marker = ''
            if show_cluster_center_marker:
                marker = '.'

            for j in range(K):
                plt.plot(
                    range(cluster_centers.shape[0]), 
                    cluster_centers[:, j, f], 
                    color=cmap(norm(j)), 
                    label=cluster_labels[j],
                    marker=marker
                    )

        ax.set_xlabel(xlabel)

        ax.set_ylabel(ylabel)
        
        ax.set_title(title_list[f])
        
        ax.set_xticks(ticks=list(range(T))) # needed so that xticks wouldn't be float
        if show_all_xticklabels:
            ax.set_xticklabels(label_dict['T'], rotation=x_rotation)  

        if f != F-1:
            ax.set_xlabel('')
            ax.set_xticklabels('')
        
        if cluster_centers is not None or labels is not None:
            plt.legend()

        axes.append(ax)

    return fig, axes

def waterfall_plot(
        time_series: npt.NDArray[np.float64],
        label_dict: dict|None = None,
        *,
        xlabel: str = 'time-axis',
        ylabel: str = 'Features-axis',
        zlabel: str = 'Feature Values',
        title: str|None = None
        ) -> Tuple[matplotlib.figure.Figure, mpl_toolkits.mplot3d.axes3d.Axes3D]:
    
    """
    Function to plot a waterfall plot of a single time series data. This data can be a time series of a single entity or cluster center.
    To make the plot interactive, use a suitable matplotlib's magic command. E.g. `%matplotlib widget`. See this site for more: https://matplotlib.org/stable/users/explain/figure/interactive.html

    Parameters
    ----------
    time_series : numpy array
        The time series data to plot. This data can be a time series of a single entity or cluster center.
        Data should be a 2-D array of shape (T, F), where T and F are the number of timesteps and features respectively.
    label_dict : dict, default=None
        a dictionary whose keys are 'T', 'N', and 'F' (which are the number of time steps, entities, and features respectively). Value of each key is a list such that the value of key:
        - 'T' is a list of names/labels of each time step to be used as index of each dataframe. If None, range(0, T) is used. Where T is the number of time steps in the fitted data
        - 'N' (ignored) is a list of names/labels of each entity. If None, range(0, N) is used. Where N is the number of entities/observations in the fitted data 
        - 'F' is a list of names/labels of each feature to be used as column of each dataframe. If None, range(0, F) is used. Where F is the number of features in the fitted data 
        If label_dict is not None, it is used to label timestep and features in the plot. 
    xlabel : str, default='time-axis'
        The label of the x-axis (time axis)
    ylabel : str, default='Features-axis'
        The label of the y-axis (feature axis)
    zlabel : str, default='Feature Values'
        The label of the z-axis (axis for the values of the features)
    title : str, default=None
        The title of the plot

    Returns
    -------
    matplotlib.figure.Figure
        the matplotlib figure object
    mpl_toolkits.mplot3d.axes3d.Axes3D
        the matplotlib's 3-D axes object. 
    """

    x = np.arange(time_series.shape[0]) # timesteps
    y = np.arange(time_series.shape[1]) # features
    
    X, Y = np.meshgrid(x, y)

    X, Y = X.T, Y.T

    # Creating a 3D plot
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    
    # Plotting the basic 3D surface
    ax.plot_surface(X, Y, time_series, color='grey', alpha=0.9)
    
    T, F = time_series.shape
    
    for f in range(F):
        ax.plot(x, [f]*T, time_series[:, f], color='red')
    
    label_dict_init = {'T': T, 'F': F}

    if label_dict is None:
        label_dict = {}
    
    for key, val in label_dict_init.items():
            _ = label_dict.setdefault(key, list(range(1, val+1)))

    x_tick_labels = label_dict['T']
    y_tick_labels = label_dict['F']

    if title is None:
        title = ''

    # Customizing the plot
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_zlabel(zlabel)
    ax.set_title(title)
    ax.set_xticks(x, x_tick_labels)
    ax.set_yticks(y, y_tick_labels)
    
    # Create sliders for elevation and azimuth angles
    ax_elev = plt.axes([0.1, 0.05, 0.8, 0.02])
    ax_azim = plt.axes([0.1, 0.02, 0.8, 0.02])
    
    def _update(val):
        ax.view_init(elev=slider_elev.val, azim=slider_azim.val)
        fig.canvas.draw_idle()
        
    slider_elev = Slider(ax_elev, 'Elevation', 0, 360, valinit=45)
    slider_azim = Slider(ax_azim, 'Azimuth', 0, 360, valinit=45)
    
    # Connect sliders to update function
    slider_elev.on_changed(_update)
    slider_azim.on_changed(_update)
    
    # Displaying the plot
    return fig, ax


from __future__ import annotations
from typing import List, Tuple

import numpy.typing as npt
import numpy as np
import pandas as pd 

from tscluster.preprocessing.utils import broadcast_data, tnf_to_ntf

class TSCluster():
    """
    Class that contains implementations of common methods for a temporal clustering model
    """


    @property
    def label_dict_(self) -> dict:
        """
        returns a dictionary of the labels whose keys are 'T', 'N', and 'F' (which are the number of time steps, entities, and features respectively). Value of each key is a list such that the value of key:
            - 'T' is a list of names/labels of each time step used as index of each dataframe during fit. Default is range(0, T). Where T is the number of time steps in the fitted data
            - 'N' is a list of names/labels of each entity used as index of the dataframe. Default is range(0, N). Where N is the number of entities/observations in the fitted data 
            - 'F' is a list of names/labels of each feature used as column of each dataframe. Default is range(0, F). Where F is the number of features in the fitted data 

        """
        keys = ('T', 'N', 'F')
        if self._label_dict_ is None:
            self._label_dict_ = {}
        
        for i, k in enumerate(keys):
            _ = self._label_dict_.setdefault(k, list(range(self.fitted_data_shape_[i])))

        return self._label_dict_
    
    def set_label_dict(self, value: dict) -> None:
        """
        Method to manually set the label_dict_.

        Parameters
        ----------
        value : dict
            the value to set as label_dict_. Should be a dict with all of 'T', 'N', and 'F' (case sensitive, which are number of time steps, entities, and features respectively) as key. The value of each key is a list of labels for the key in the data.  If your data don't have values for any of the keys, set its value to None.
        
        Returns
        -------
        dict 
            a dictionary whose keys are 'T', 'N', and 'F'; and values are lists of the labels of each key.
        """

        valid_keys = {'T', 'N', 'F'}

        if not isinstance(value, dict):
            raise TypeError(f"Expected value to be of type 'dict', but got '{type(value).__name__}'")
        elif any(k not in valid_keys for k in value.keys()):
            raise ValueError(f"Expected dict to have all of {valid_keys} as key.")
        
        self._label_dict_ = value


    def get_named_cluster_centers(
                                self, 
                                label_dict : dict|None = None
                                   ) -> List[pd.DataFrame]:
        """
        Method to return the cluster centers with custom names of time steps and features.

        Parameters
        ----------
        label_dict dict, default=None
            a dictionary whose keys are 'T', 'N', and 'F' (which are the number of time steps, entities, and features respectively). Value of each key is a list such that the value of key:
            - 'T' is a list of names/labels of each time step to be used as index of each dataframe. If None, range(0, T) is used. Where T is the number of time steps in the fitted data
            - 'N' is a list of names/labels of each entity to be used as index of the dataframe. If None, range(0, N) is used. Where N is the number of entities/observations in the fitted data 
            - 'F' is a list of names/labels of each feature to be used as column of each dataframe. If None, range(0, F) is used. Where F is the number of features in the fitted data 
            If label_dict is None, the result of self.label_dict_ is used.

        Returns
        ------ 
        list    
            A list of k pandas DataFrames. Where k is the number of clusters. The i-th dataframe in the list is a T x F dataframe of the values of the cluster centers of the i-th cluster.   
        """

        cluster_centers, _ = broadcast_data(self.fitted_data_shape_[0], cluster_centers=self.cluster_centers_)

        cluster_centers = tnf_to_ntf(cluster_centers)
    
        if label_dict is None:
            time, features = self.label_dict_['T'], self.label_dict_['F']

        else:
            time, features = label_dict['T'], label_dict['F']

            if time is None:
                time = list(range(self.fitted_data_shape_[0]))
            if features is None:
                features = list(range(self.fitted_data_shape_[2]))

        return [pd.DataFrame(cluster_centers_k, columns=features, index=time) for cluster_centers_k in cluster_centers]

    def get_named_labels(
                        self, 
                        label_dict: dict|None = None
                        ) -> pd.DataFrame:
        """
        Method to return the a data frame of the label assignments with custom names of time steps and entities.

        Parameters
        -----------
        label_dict : dict, default=None
            a dictionary whose keys are 'T', 'N', and 'F' (which are the number of time steps, entities, and features respectively). Value of each key is a list such that the value of key:
            - 'T' is a list of names/labels of each time step to be used as index of each dataframe. If None, range(0, T) is used. Where T is the number of time steps in the fitted data
            - 'N' is a list of names/labels of each entity to be used as index of the dataframe. If None, range(0, N) is used. Where N is the number of entities/observations in the fitted data 
            - 'F' is a list of names/labels of each feature to be used as column of each dataframe. If None, range(0, F) is used. Where F is the number of features in the fitted data 
            If label_dict is None, the result of self.label_dict_ is used.

        Returns
        -------
        pd.DataFrame
            A pandas DataFrame with shape (N, T). The value in the n-th row and t-th column is an integer indicating the custer assignment of the n-th entity/observation at time t.  
        """
        
        _, labels = broadcast_data(self.fitted_data_shape_[0], labels=self.labels_)
    
        if label_dict is None:
            time, entities = self.label_dict_['T'], self.label_dict_['N']

        else:
            time, entities = label_dict['T'], label_dict['N']

            if time is None:
                time = list(range(self.fitted_data_shape_[0]))
            if entities is None:
                entities = list(range(self.fitted_data_shape_[1]))

        return pd.DataFrame(labels, columns=time, index=entities)

    def _get_changes(self) -> npt.NDArray[np.int64]:
        labels = self.get_named_labels().values
        return np.sum(labels[:, :-1] != labels[:, 1:], axis=1)
    
    def get_dynamic_entities(self) -> Tuple[List[np.int64], List[np.int64]]:
        """
        returns the dynamic entities and their number of changes. Both lists are sorted by the number of cluster changes in descending order.

        Returns
        -------
        dynamic entities : list
            a 1-D array of the indexes of the entities that change cluster at least once.
        number of changes : list
            a 1-D array of the number of changes for each dynamic entity such that the i-th element is the number of cluster changes for the i-th dynamic entity
        """
        
        n_changes = self._get_changes()
        changes = n_changes > 0

        entities = self.label_dict_['N']
        dynamic_entities = np.where(changes)[0]

        sort_filter = np.argsort(n_changes[changes])[::-1]

        return [entities[i] for i in dynamic_entities[sort_filter]], list(np.sort(n_changes[changes])[::-1])

    def get_index_of_label(self, labels: List[str], axis: str = 'N') -> List[int]:
        """
        function to return the integer indexes of some given labelled items in `self.label_dict_`. The indexes are assumed to be 0-indexed.

        Parameters
        ----------
        labels : list
            a list of the label(s) whose integer indexes should be returned.
        axis : str, default='N'
            can be any of {'T', 'N', 'F'}. 
            - If 'T', the values in the `labels` parameter are interpreted as time labels (as stored in `self.label_dict_['T']`). 
            - If 'N', the values in the `labels` parameter are interpreted as entity labels (as stored in `self.label_dict_['N']`). 
            - If 'F', the values in the `labels` parameter are interpreted as feature labels (as stored in `self.label_dict_['F']`). 
    
        Returns
        -------
        list
            a list of the integer indexes of the labels in the given axis dimension.
        """ 

        return [self.label_dict_[axis].index(label) for label in labels]

    def get_label_of_index(self, indexes: List[int], axis: str = 'N') -> List[str]:
        """
        function to return the labels of some given integer indexes as labelled in `self.label_dict_`. The indexes are assumed to be 0-indexed.

        Parameters
        ----------
        indexes : list
            a list of the index(es) whose labels should be returned.
        axis : str, default='N'
            can be any of {'T', 'N', 'F'}. 
            - If 'T', the values in the `indexes` parameter are interpreted as the time indexes whose labels (as stored in `self.label_dict_['T']`) should be returned. 
            - If 'N', the values in the `indexes` parameter are interpreted as the entity indexes whose labels (as stored in `self.label_dict_['N']`) should be returned. 
            - If 'F', the values in the `indexes` parameter are interpreted as the feature indexes whose labels (as stored in `self.label_dict_['F']`) should be returned. 
    
        Returns
        -------
        list
            a list of the labels of the given integer indexes in the given axis dimension.
        """ 
        
        return list(pd.Series(self.label_dict_[axis]).values[indexes])
    
    @property
    def n_changes_(self) -> int:
        """
        returns the total number of label changes
        """
        return np.sum(self._get_changes())


from __future__ import annotations
from typing import List, Tuple
from abc import ABC, abstractmethod

import numpy.typing as npt
import numpy as np
import pandas as pd 

class TSClusterInterface(ABC):
    @abstractmethod
    def fit(self, 
            X: npt.NDArray[np.float64], 
            label_dict: dict|None = None
            ) -> 'TSClusterInterface':
        """
        Fit method of model. 

        Parameters
        ----------
        X : ndarray
            Input time series data. Should be a 3 dimensional array in TNF fromat.
        label_dict : dict, default=None
            A dictionary of the labels of X. Keys should be 'T', 'N', and 'F' (which are the number of time steps, entities, and features respectively). Value of each key is a list such that the value of key:
                - 'T' is a list of names/labels of each time step used as index of each dataframe during fit. Default is range(0, T). Where T is the number of time steps in the fitted data
                - 'N' is a list of names/labels of each entity used as index of the dataframe. Default is range(0, N). Where N is the number of entities/observations in the fitted data 
                - 'F' is a list of names/labels of each feature used as column of each dataframe. Default is range(0, F). Where F is the number of features in the fitted data 

            data_loader function from tscluster.preprocessing.utils can help in getting label_dict of a data. 
            
        Returns
        -------
        self
            The fitted model object
        """
        
        raise NotImplementedError
    
    @property
    @abstractmethod
    def cluster_centers_(self) -> npt.NDArray[np.float64]:
        """
        returns the cluster centers. If scheme is fixed centers, returns a k x F 2D array. Where k is the number of clusters and F is the number of features. If scheme is changing centers, returns a T x k x F 3D array. Where T is the number of time stesp, k is the number of clusters and F is the number of features.
        """

        raise NotImplementedError
    
    @property
    @abstractmethod
    def labels_(self) -> npt.NDArray[np.float64]:
        """
        returns the assignment labels. values are integers in range [0, k-1], where k is the number of clusters. If scheme is fixed assignment, returns a 1D array of size N. Where N is the number of entities. A value of j at the i-th index means that entity i is assigned to the j-th cluster at all time steps. If scheme is changing assignment, returns a N x T 2D array. Where N is the number of entities and T is the number of time steps. A value of j at the i-th row and t-th column means that entity i is assigned to the j-th cluster at the t-th time step.
        """
        
        raise NotImplementedError

    @property
    @abstractmethod
    def fitted_data_shape_(self) -> Tuple[int, int, int]:
        """
        returns a tuple of the shape of the fitted data in TNF format. E.g (T, N, F) where T, N, and F are the number of timesteps, observations, and features respectively. 
        """
        
        raise NotImplementedError

    @property
    @abstractmethod
    def label_dict_(self) -> dict:
        """
        returns a dictionary of the labels whose keys are 'T', 'N', and 'F' (which are the number of time steps, entities, and features respectively). Value of each key is a list such that the value of key:
            - 'T' is a list of names/labels of each time step used as index of each dataframe during fit. Default is range(0, T). Where T is the number of time steps in the fitted data
            - 'N' is a list of names/labels of each entity used as index of the dataframe. Default is range(0, N). Where N is the number of entities/observations in the fitted data 
            - 'F' is a list of names/labels of each feature used as column of each dataframe. Default is range(0, F). Where F is the number of features in the fitted data 
        """

        raise NotImplementedError
    
    @abstractmethod
    def set_label_dict(self, value: dict) -> None:
        """
        Method to manually set the label_dict_.

        Parameters
        ----------
        value : dict
            the value to set as label_dict_. Should be a dict with all of 'T', 'N', and 'F' (case sensitive, which are number of time steps, entities, and features respectively) as key. The value of each key is a list of labels for the key in the data.  If your data don't have values for any of the keys, set its value to None.
        """

        raise NotImplementedError

    @abstractmethod
    def get_named_cluster_centers(
                                self, 
                                label_dict: dict|None = None
                                   ) -> List[pd.DataFrame]:
        """
        Method to return the cluster centers with custom names of time steps and features.

        Parameters
        -----------
        label_dict dict, default=None
            a dictionary whose keys are 'T', 'N', and 'F' (which are the number of time steps, entities, and features respectively). Value of each key is a list such that the value of key:
            - 'T' is a list of names/labels of each time step to be used as index of each dataframe. If None, range(0, T) is used. Where T is the number of time steps in the fitted data
            - 'N' is a list of names/labels of each entity to be used as index of the dataframe. If None, range(0, N) is used. Where N is the number of entities/observations in the fitted data 
            - 'F' is a list of names/labels of each feature to be used as column of each dataframe. If None, range(0, F) is used. Where F is the number of features in the fitted data 
            If label_dict is None, the result of self.label_dict_ is used.

        Returns
        --------
        list
            A list of k pandas DataFrames. Where k is the number of clusters. The i-th dataframe in the list is a T x F dataframe of the values of the cluster centers of the i-th cluster.   
        """
        
        raise NotImplementedError 

    @abstractmethod
    def get_named_labels(
                        self, 
                        label_dict: dict|None = None
                        ) -> pd.DataFrame:
        """
        Method to return the a data frame of the label assignments with custom names of time steps and entities.

        Parameters
        -----------
        label_dict dict, default=None
            a dictionary whose keys are 'T', 'N', and 'F' (which are the number of time steps, entities, and features respectively). Value of each key is a list such that the value of key:
            - 'T' is a list of names/labels of each time step to be used as index of each dataframe. If None, range(0, T) is used. Where T is the number of time steps in the fitted data
            - 'N' is a list of names/labels of each entity to be used as index of the dataframe. If None, range(0, N) is used. Where N is the number of entities/observations in the fitted data 
            - 'F' is a list of names/labels of each feature to be used as column of each dataframe. If None, range(0, F) is used. Where F is the number of features in the fitted data 
            If label_dict is None, the result of self.label_dict_ is used.

        Returns
        -------- 
        pd.DataFrame
            A pandas DataFrame with shape (N, T). The value in the n-th row and t-th column is an integer indicating the custer assignment of the n-th entity/observation at time t.  
        """
        
        raise NotImplementedError 


from __future__ import annotations
from typing import List

import numpy as np
import numpy.typing as npt
import pandas as pd
from sklearn.preprocessing import StandardScaler, MinMaxScaler

from tscluster.preprocessing.interface import TSPreprocessorInterface
from tscluster.preprocessing.utils import reshape_for_transform, infer_data

class TSScaler(TSPreprocessorInterface):
    def __init__(self, scaler, per_time: bool = True, **kwargs) -> None:
        # parent class for transformers

        self._scaler = scaler # scaler object (e.g. sklearn's scaler obejct for each time step)
        self.per_time = per_time
        self.kwargs = kwargs 
    
    @property
    def label_dict_(self) -> dict:
        """
        returns a dictionary of the labels. Keys are: 'T', 'N', and 'F' which are the number of time steps, entities, and features respectively. Each key's value is a list of its labels seen during the last time the data was fitted, transformed or inverse_transformed. 
        """
        return self._label_dict_
    
    def set_label_dict_(self, value: dict) -> None:
        """
        Method to manually set the label_dict_.

        Parameters
        ----------
        value : dict
            the value to set as label_dict_. Should be a dict with all of 'T', 'N', and 'F' (case sensitive, which are number of time steps, entities, and features respectively) as key. The value of each key is a list of labels for the key in the data.  If your data don't have values for any of the keys, set its value to None.
        """

        valid_keys = {'T', 'N', 'F'}

        if not isinstance(value, dict):
            raise TypeError(f"Expected value to be of type 'dict', but got '{type(value).__name__}'")
        elif any(k not in valid_keys for k in value.keys()):
            raise ValueError(f"Expected dict to have all of {valid_keys} as key.")
        
        self._label_dict_ = value

    def fit(
        self, 
        X: npt.NDArray[np.float64],
        **kwargs
        ) -> 'TSScaler':
        """
        Fit method of transformer. 

        Parameters
        ----------
        X: ndarray
            Input time series data. Should be a 3 dimensional array in TNF fromat.
       **kwargs keyword arguments to be passed to fit method.
            
        Returns 
        -------
        self
            the fitted transformer object
        """
        
        X, n = reshape_for_transform(X, self.per_time)

        self._scalers = [self._scaler(**self.kwargs).fit(X[i], *kwargs) for i in range(n)] 

        return self
    
    def transform(
        self, 
        X: npt.NDArray[np.float64], 
        **kwargs
        ) -> npt.NDArray[np.float64]:
        """
        transform method for  transformer. 

        Parameters
        ----------
        X: ndarray
            Input time series data. Should be a 3 dimensional array in TNF fromat.
       **kwargs keyword arguments.

        Returns
        -------
        numpy array 
            the transformed data in TNF format
        """
       
        _shape = X.shape

        X, _ = reshape_for_transform(X, self.per_time)

        return np.array([scaler.transform(X[i]) for i, scaler in enumerate(self._scalers)]).reshape(*_shape)
    
    def inverse_transform(
        self, 
        X: npt.NDArray[np.float64],
        **kwargs
        ) -> npt.NDArray[np.float64]:
        """
        inverse transform method for  transformer. 

        Parameters
        ----------
        X: ndarray. 
            Input time series data. Should be a 3 dimensional array in TNF fromat.
       **kwargs keyword arguments.
    
        Returns
        -------
        numpy array 
            the inverse-transform of the data in TNF format
        """

        _shape = X.shape

        X, _ = reshape_for_transform(X, self.per_time)

        return np.array([scaler.inverse_transform(X[i]) for i, scaler in enumerate(self._scalers)]).reshape(*_shape)

    def fit_transform(
        self, 
        X: npt.NDArray[np.float64],
        **kwargs
        ) -> npt.NDArray[np.float64]:
        """
        fit and transform the data

        Parameters
        ----------
        X: ndarray
            Input time series data. Should be a 3 dimensional array in TNF fromat.
       **kwargs keyword arguments
       
        Returns
        -------
        numpy array 
            the transformed data in TNF format
        """

        self.fit(X)

        return self.transform(X)

class TSStandardScaler(TSScaler):
    """
    Uses sklearn's StandardScaler to scale a time series data.

    Parameters
    -----------
    per_time : bool, default=True
        If True, compute zscore per time step. If False, compute zscore per feature across all timesteps.
        
    **kwargs : keyword arugments to be passed to sklearn's StandardScaler.
    """

    def __init__(self, per_time: bool = True, **kwargs) -> None:

        scaler = StandardScaler
        super().__init__(scaler, per_time, **kwargs)

class TSMinMaxScaler(TSScaler):
    """
    Uses sklearn's MinMaxScaler to scale a time series data.

    Parameters
    -----------
    per_time : bool, default=True
        If True, compute zscore per time step. If False, compute zscore per feature across all timesteps.
        
    **kwargs : keyword arugments to be passed to sklearn's MinMaxScaler.
    """

    def __init__(self, per_time: bool = True, **kwargs) -> None:

        scaler = MinMaxScaler
        super().__init__(scaler, per_time, **kwargs)



from __future__ import annotations
from typing import List, Any, Tuple

import numpy as np
import numpy.typing as npt
from tslearn.clustering import TimeSeriesKMeans

from tscluster.interface import TSClusterInterface
from tscluster.base import TSCluster
from tscluster.preprocessing.utils import tnf_to_ntf, ntf_to_tnf, infer_data

class TSKmeans(TimeSeriesKMeans, TSCluster, TSClusterInterface):
    """
    Applies tslearn's TimeSeriesKMeans clustering to the data. 

    Read more in the tslearn's user guide. The follow parameters are from tslearn's TimeSeriesKMeans constructor.

    Parameters
    ----------
    n_clusters : int (default: 3)
        Number of clusters to form.

    max_iter : int (default: 50)
        Maximum number of iterations of the k-means algorithm for a single run.

    tol : float (default: 1e-6)
        Inertia variation threshold. If at some point, inertia varies less than this threshold between two consecutive iterations, the model is considered to have converged and the algorithm stops.

    n_init : int (default: 1)
        Number of time the k-means algorithm will be run with different centroid seeds. The final results will be the best output of n_init consecutive runs in terms of inertia.

    metric : {"euclidean", "dtw", "softdtw"} (default: "euclidean")
        Metric to be used for both cluster assignment and barycenter computation. If "dtw", DBA is used for barycenter computation.

    max_iter_barycenter : int (default: 100)
        Number of iterations for the barycenter computation process. Only used if metric="dtw" or metric="softdtw".

    metric_params : dict or None (default: None)
        Parameter values for the chosen metric. For metrics that accept parallelization of the cross-distance matrix computations, n_jobs key passed in metric_params is overridden by the n_jobs argument.

    n_jobs : int or None, optional (default=None)
        The number of jobs to run in parallel for cross-distance matrix computations. Ignored if the cross-distance matrix cannot be computed using parallelization.
        None means 1 unless in a joblib.parallel_backend context. -1 means using all processors. See scikit-learns'     Glossary <https://scikit-learn.org/stable/glossary.html#term-n-jobs>_ for more details.

    dtw_inertia: bool (default: False)
        Whether to compute DTW inertia even if DTW is not the chosen metric.

    verbose : int (default: 0)
        If nonzero, print information about the inertia while learning the model and joblib progress messages are printed.

    random_state : integer or numpy.RandomState, optional
        Generator used to initialize the centers. If an integer is given, it fixes the seed. Defaults to the global numpy random number generator.

    init : {'k-means++', 'random' or an ndarray} (default: 'k-means++')
        Method for initialization:
        'k-means++' : use k-means++ heuristic. See scikit-learn's k_init_ <https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/ cluster/k_means_.py>_ for more.
        'random': choose k observations (rows) at random from data for the initial centroids. If an ndarray is passed, it should be of shape (n_clusters, ts_size, d) and gives the initial centers.
    
    Attributes
    ----------
    cluster_centers_
    fitted_data_shape_
    labels_
    label_dict_
    n_changes_
    """
    
    def fit(
        self, 
        X: npt.NDArray[np.float64],
        label_dict: dict|None = None, 
        **kwargs
        ) -> 'TSKmeans':
        """
        Method for fitting the model on the data.

        Parameters
        ----------
        X : numpy array
            Input time series data.Should be a 3 dimensional array in TNF fromat.
        label_dict : dict, default=None
            A dictionary of the labels of X. Keys should be 'T', 'N', and 'F' (which are the number of time steps, entities, and features respectively). Value of each key is a list such that the value of key:
                - 'T' is a list of names/labels of each time step used as index of each dataframe during fit. Default is range(0, T). Where T is the number of time steps in the fitted data
                - 'N' is a list of names/labels of each entity used as index of the dataframe. Default is range(0, N). Where N is the number of entities/observations in the fitted data 
                - 'F' is a list of names/labels of each feature used as column of each dataframe. Default is range(0, F). Where F is the number of features in the fitted data 
            data_loader function from tscluster.preprocessing.utils can help in getting label_dict of a data. 
        **kwargs keyword arguments to be passed to tslearn's kmeans fit method

        Returns
        --------
        self: 
            The fitted TSKmeans object.
        """

        self._label_dict_ = label_dict

        self._labels_ = None
        self._cluster_centers_ = None

        self.Xt = tnf_to_ntf(X)

        self.N_, self.T_, self.F_ = self.Xt.shape

        super().fit(self.Xt, **kwargs) 

        self._cluster_centers_ = ntf_to_tnf(self._cluster_centers_)

        return self

    @property
    def cluster_centers_(self) -> npt.NDArray[np.float64]: 
        return self._cluster_centers_
    
    @cluster_centers_.setter
    def cluster_centers_(self, new_value: npt.NDArray[np.float64]) -> None:
        self._cluster_centers_ = new_value

    @property
    def labels_(self) -> npt.NDArray[np.int64]:
        """
        Returns the assignment labels. Values are integers in range [0, k-1], where k is the number of clusters. If scheme is fixed assignment, returns a 1D array of size N. Where N is the number of entities. A value of j at the i-th index means that entity i is assigned to the j-th cluster at all time steps. If scheme is changing assignment, returns a N x T 2D array. Where N is the number of entities and T is the number of time steps. A value of j at the i-th row and t-th column means that entity i is assigned to the j-th cluster at the t-th time step.
        """
        return self._labels_ 
    
    @labels_.setter
    def labels_(self, new_value: npt.NDArray[np.int64]) -> None:
        self._labels_ = new_value

    @property
    def fitted_data_shape_(self) -> Tuple[int, int, int]:
        """
        returns a tuple of the shape of the fitted data in TNF format. E.g (T, N, F) where T, N, and F are the number of timesteps,
        observations, and features respectively. 
        """
        return self.T_, self.N_, self.F_


from __future__ import annotations
import copy
from time import time
from typing import Tuple, List

import numpy as np
import numpy.typing as npt
from scipy.cluster.vq import kmeans, vq
import gurobipy as gp

class MILPClustering():
    def __init__(
            self, 
            k: int, 
            IFrac: float = 0.2, 
            use_full_constraints: bool = False, 
            max_iter: int = 10, 
            add_constaint_per_cluster: bool = True, 
            random_state: None|int = None, 
            warm_start: bool = True, 
            verbose: bool = True, 
            top_n_percentile: float = 0.0
            ) -> None:
        
        self.k = k
        self.IFrac = IFrac
        self.random_state = random_state
        self.warm_start = warm_start
        self.verbose = verbose
        self.use_full_constraints = use_full_constraints
        self.max_iter = max_iter
        self.add_constaint_per_cluster = add_constaint_per_cluster
        self.top_n_percentile = top_n_percentile 

        self.Es_star = None
        self.Zs_ans = None
        self.Cs_hats = None
        self.I_adds = None
        self.zis = None
        self.Z = None
        self.warm_start_labels_ = None

        self.x_min = 0
    
    def fit(
            self, 
            X: npt.NDArray[np.float64] | npt.NDArray[np.int64], 
            y: npt.NDArray[np.float64] | npt.NDArray[np.int64] | None = None
            ) -> "MILPClustering":
        
        X = copy.deepcopy(X)

        n = len(X)

        if np.min(X) < 0:
            self.x_min = np.min(X) - 1
            X -= self.x_min
            
        # inital random X
        if self.random_state is not None:
            np.random.seed(self.random_state)

        I_idx = list(np.random.choice(list(range(n)), size=np.int64(np.round(self.IFrac*n)), replace=False))
        
        if self.use_full_constraints:
            I = X 
        else:
            I = X[I_idx, :]
    
        if self.warm_start:
            # warm start clusters
            km_Z, _ = kmeans(X, self.k)

            km_labels, _ = vq(X, km_Z)

            km_Z_to_Z_mapper = km_Z[:, 0].argsort()

            Z = km_Z[km_Z_to_Z_mapper, :]

            C = np.zeros((I.shape[0], self.k))

            km_Z_to_Z_mapper_key_val_pairs = np.array(list(enumerate(km_Z_to_Z_mapper)))

            for i in range(C.shape[0]):
                j = km_Z_to_Z_mapper_key_val_pairs[:, 0][km_Z_to_Z_mapper_key_val_pairs[:, 1] == km_labels[i]] 
                C[i, j] = 1        
                
        E_star = 0
        E_hat = np.Inf
        count = 1
        
        epsilon = 1e-4
        
        Es_star = []
        Zs_ans = []
        Cs_hats = []
        I_adds = []
        zis = []
        
        t0 = time() 

        while E_hat > E_star+epsilon and count < self.max_iter+1:
            # solve MILP
            if self.warm_start:
                E_star, Z_ans, C_ans = MILPClustering.solve_MILP(I, self.k, Z, C)
            else:
                E_star, Z_ans, C_ans = MILPClustering.solve_MILP(I, self.k)
                
            C_hat = MILPClustering.assign_cluster(X, Z_ans)

            E_hat, I_add, zi = MILPClustering.add_constraints(X, Z_ans, C_hat, per_cluster=self.add_constaint_per_cluster,
                                                              top_n_percentile=self.top_n_percentile)
                    
            I_idx = set(I_idx) 
            I_idx.update(I_add)
            
            I_idx = list(I_idx)
            
            I = X[I_idx, :]
            
            if self.warm_start:
                C = np.zeros((I.shape[0], self.k))
                for i in range(C.shape[0]):
                    j = km_Z_to_Z_mapper_key_val_pairs[:, 0][km_Z_to_Z_mapper_key_val_pairs[:, 1] == km_labels[i]] 
                    C[i, j] = 1
            
            if self.verbose:    
                print(f"Done with {count} of {self.max_iter}. Ehat: {E_hat}, Estar: {E_star}")
        
                count += 1
            
            Es_star.append(E_star)
                        
            Zs_ans.append(Z_ans + self.x_min)
            Cs_hats.append(C_hat)
            I_adds.append(I_add)
            zis.append(np.array(zi) + self.x_min)

        self.Es_star = Es_star
        self.Zs_ans = Zs_ans
        self.Cs_hats = Cs_hats
        self.I_adds = I_adds
        self.zis = zis
        
        if self.warm_start:
            self.Z = Z + self.x_min
            self.warm_start_labels_ = km_labels

        if self.verbose:
            print(f"Total time is {np.round(time() - t0, 2)}secs")  
            print()

        return self

    def predict(
            self, 
            X: npt.NDArray[np.float64] | npt.NDArray[np.int64]
            ) -> npt.NDArray[np.float64] | npt.NDArray[np.int64]:
        
        return np.argmax(MILPClustering.assign_cluster(X, self.Z_ans), axis=1)

    @staticmethod
    def solve_MILP(
        I: npt.NDArray[np.float64] | npt.NDArray[np.int64], 
        k: int, 
        init_Z: npt.NDArray[np.float64] | npt.NDArray[np.int64] | None = None, 
        init_C: npt.NDArray[np.float64] | npt.NDArray[np.int64] | None = None
        ) -> Tuple[float, npt.NDArray[np.float64], npt.NDArray[np.float64]]:

        # init_Z is warm start
        
        # MILP
        m = gp.Model()
        
        # creating Vars
        E = m.addVar(name='E')
        
        Z_vars = m.addVars(k, I.shape[1], lb=0, ub=gp.GRB.INFINITY, vtype=gp.GRB.CONTINUOUS)
        if init_Z is not None:
            for zi in range(k):
                for zj in range(I.shape[1]):
                    Z_vars[zi, zj].start = init_Z[zi, zj]
                
        C_vars = m.addVars(I.shape[0], k, vtype=gp.GRB.BINARY)
        if init_C is not None:
            for ci in range(I.shape[0]):
                for cj in range(k):
                    C_vars[ci, cj].start = init_C[ci, cj]
                
        e = m.addVars(I.shape[0], k, I.shape[1]) # to track absolute difference

        # creating Contraints
        for i in range(I.shape[0]):
            for j in range(k):
                for f in range(I.shape[1]):
    #                 e[i, j, f] = m.addVar(name='e_'+str(i)+'_'+str(j)+'_'+str(f))
                
                    m.addConstr(e[i, j, f] >= I[i, f] - Z_vars[j, f])
                    m.addConstr(e[i, j, f] >= -(I[i, f] - Z_vars[j, f]))
                    
        for i in range(I.shape[0]):
            for j in range(k):
                m.addConstr((C_vars[i, j] == 1) >> (E >= gp.quicksum(e[i, j, f] for f in range(I.shape[1]))))
                
            m.addConstr(gp.quicksum(C_vars[i, j] for j in range(k)) == 1)
            
        eps = 1e-3
        for j in range(k - 1):
            m.addConstr(Z_vars[j, 0] + eps <= Z_vars[j+1, 0])
        
        m.setObjective(E, gp.GRB.MINIMIZE)
        m.setParam('OutputFlag', 0)
        
        m.optimize() 
        
        Z_ans = np.zeros((k, I.shape[1]))
        C_ans = np.zeros((I.shape[0], k))

        for ji in range(k):
            for jj in range(I.shape[1]):
                Z_ans[ji, jj] = Z_vars[ji, jj].X

            for ii in range(I.shape[0]):
                C_ans[ii, ji] = C_vars[ii, ji].X
                
        return E.X, Z_ans, C_ans

    @staticmethod
    def assign_cluster(
        X: npt.NDArray[np.float64] | npt.NDArray[np.int64], 
        Z: npt.NDArray[np.float64] | npt.NDArray[np.int64]
        ) -> npt.NDArray[np.float64] | npt.NDArray[np.int64]:
    
        Z_cluster = []
            
        for z in range(Z.shape[0]):
            Z_cluster.append(np.linalg.norm(X - Z[z, :], ord=1, axis=1))
            
        C_hat = np.zeros((X.shape[0], Z.shape[0]))
        
        C_hat[list(range(X.shape[0])), np.argmin(np.array(Z_cluster), axis=0)] = 1
        
        return C_hat

    @staticmethod
    def add_constraints(
        X: npt.NDArray[np.float64] | npt.NDArray[np.int64], 
        Z: npt.NDArray[np.float64] | npt.NDArray[np.int64], 
        C_hat: npt.NDArray[np.float64] | npt.NDArray[np.int64], 
        per_cluster: bool = False, 
        top_n_percentile: float = 0.0
        ) -> Tuple[float, List, List]:

        E_hat = 0.0

        I_add = []
        zi = []

        if top_n_percentile:
            top_n_percentile_cut = np.int64(np.round(top_n_percentile*X.shape[0]))
        else:
            top_n_percentile_cut = 1

        # solve for E_hat = max violation
        for z in range(Z.shape[0]):
            distance = np.linalg.norm(X - Z[z, :], ord=1, axis=1) * C_hat[:, z]
            
            max_distance = np.max(distance)
            
            # optimize way of finding max Ehat during iteration
            if E_hat < max_distance:
                E_hat = max_distance

                if not per_cluster:
                    I_add = np.argsort(distance)[::-1][:top_n_percentile_cut]  #[np.argmax(distance) ]
                    zi = [z] 

            if per_cluster:
                I_add.extend(np.argsort(distance)[::-1][:top_n_percentile_cut])  #.append(np.argmax(distance))
                zi.append(z) 

        return E_hat, I_add, zi

if __name__ == "__main__":
    pass



from setuptools import setup, find_packages

with open("README.md", 'r') as f:
    long_description = f.read()

setup(
   name='tscluster',
   version='1.0.4',
   description='A useful package for temporal clustering',
   license="MIT",
   long_description=long_description,
   long_description_content_type="text/markdown",
   author='Jolomi Tosanwumi',
   author_email='tjolomi@gmail.com',
   #url="...",
   packages=find_packages(), 
   #python_requires='>=3.9.18',
   install_requires=[
       'numpy>=1.26', 
       'scipy>=1.10', 
       'gurobipy>=11.0', 
       'tslearn>=0.6.3',   
       'h5py>=3.10',
       'pandas>=2.2',
       'matplotlib>=3.8,<3.9'
       ], #external packages as dependencies on TEST pypi
   #ToDo: on real pypi: ['numpy==1.26.4', 'scipy==1.12.0', 'gurobipy==11.0.1']. Also considering using >= a lowerbound
)